"""
@author: Bert Coerver
"""
import numpy as np
from numba import types, guvectorize, njit, float64, int64

def second_order_diff_matrix(x):
    X = np.append(x, [np.nan, np.nan])
    # Create x-aware delta matrix. When sample-points are at equal distance,
    # this reduces to [[1, -2, 1, ...], ..., [..., 1, -2, 1]].
    diag_n0 = 2  / ((X[1:-1] - X[:-2])  * (X[2:]   - X[:-2]))
    diag_n1 = -2 / ((X[2:-1] - X[1:-2]) * (X[1:-2] - X[0:-3]))
    diag_n2 = 2  / ((X[2:-2] - X[1:-3]) * (X[2:-2] - X[0:-4]))
    D = np.diag(diag_n0) + np.diag(diag_n1, k = 1) + np.diag(diag_n2, k = 2)
    D = D[:-2,:]
    A = np.dot(D.T, D)
    return A

@njit(types.Tuple((float64[:], float64, float64))(float64[:], float64[::1], float64, float64[:], float64[:], float64, float64[:,::1], float64))
def iterator(Z, Y, a_, w, u, lmb, A, a):
    v = np.where(Y > Z, a_, 1 - a_)
    W = np.diag(v*w*u)
    Z_ = np.linalg.solve(W + lmb * A, np.dot(W,Y))
    rmse = np.sqrt(np.mean((Z - Z_)**2))
    a_ = a
    return Z_, rmse, a_

@njit(types.Tuple((float64[:], float64[::1], float64[:], float64, int64, float64))(float64[:]))
def initiator(Y):
    w = np.isfinite(Y, np.zeros_like(Y))
    Y_ = np.where(w == 0, 0, Y)
    a_ = rmse = 0.5
    j = 0
    Z = np.zeros_like(Y)
    return Z, Y_, w, a_, j, rmse

@njit(float64[:](float64[::1], float64[:,::1], float64, float64[:], float64, float64, float64, int64))
def whittaker(Y, A, lmbda, u, a, min_drange, max_drange, max_iter):
    Z, Y_, w, a_, j, rmse = initiator(Y)
    relevant_idxs_size = n_peaks = 0
    while j == 0 or ((not (a == 0.5 or rmse < 0.01) or (relevant_idxs_size > 0)) and (j < max_iter)):
        Z[:], rmse, a_ = iterator(Z, Y_, a_, w, u, lmbda, A, a)
        j += 1

        idxs = np.diff(np.sign(Z[1:] - Z[:-1])).nonzero()[0] + 1
        if (j == 1) or (n_peaks != idxs.size):
            z = Z[idxs]
            is_peak_valley = (z > max_drange) | (z < min_drange)
            is_valley = (z < min_drange)[is_peak_valley]
            peaks_valleys = idxs[is_peak_valley]
            peak_valley_n = np.searchsorted(idxs, peaks_valleys)
            inequal_swap = np.where(is_valley, -1, 1)
            peak_limit = np.where(is_valley, min_drange, max_drange)
            orig_z = Z[peaks_valleys]
            n_peaks = idxs.size
            m = 0

        not_solved = np.take(Z, idxs[peak_valley_n]) * inequal_swap > peak_limit * inequal_swap
        m += 1
        relevant_idxs = peaks_valleys[not_solved]
        Y_[relevant_idxs] = peak_limit[not_solved] - 0.2 * m * (orig_z[not_solved] - peak_limit[not_solved])
        w[relevant_idxs] = 1.0
        relevant_idxs_size = relevant_idxs.size
    return Z

@guvectorize([(float64[::1], float64[:,::1], float64, float64[:], float64, float64, float64, int64, float64[:])], 
             '(n),(n,n),(),(n),(),(),(),()->(n)', nopython = True, target = "parallel")
def _wt1(Y, A, lmbda, u, a, min_drange, max_drange, max_iter, Z):
    Z[:] = whittaker(Y, A, lmbda, u, a, min_drange, max_drange, max_iter)

@guvectorize([(float64[::1], float64[:,::1], float64[:], float64[:], float64, float64, float64, int64, float64[:,:])], 
             '(n),(n,n),(k),(n),(),(),(),()->(k,n)', nopython = True, target = "parallel")
def _wt2(Y, A, lmbda, u, a, min_drange, max_drange, max_iter, Z):
    for i, lmb in enumerate(lmbda):
        Z[i,:] = whittaker(Y, A, lmb, u, a, min_drange, max_drange, max_iter)

def wt(Y, A, lmbdas, u, a, min_drange, max_drange, max_iter):
    funcs = [_wt1, _wt2]
    y_dims = getattr(Y, "ndim", 0)
    lmbdas_dims = getattr(lmbdas, "ndim", 0)
    if y_dims in [2,3] and lmbdas_dims in [1]:
        wt_func = funcs[1]
    elif y_dims in [2] and lmbdas_dims in [2]:
        raise ValueError
    else:
        wt_func = funcs[0]
    Z = wt_func(Y, A, lmbdas, u, a, min_drange, max_drange, max_iter)
    return Z

def dist_to_finite(y, x, axis = -1, extrapolate = False):
    def _dtf(y, x, extrapolate = False):
        # NOTE could also use `xarray.core.missing._get_nan_block_lengths instead` (similar, but not exactly the same)
        # NOTE can use numba when this is resolved https://github.com/numba/numba/issues/1269
        part1 = np.expand_dims(np.extract(np.isfinite(y), x), 1)
        part2 = np.expand_dims(x, 0)
        distances = part2 - part1
        dist = np.min(np.abs(distances), axis = 0, initial = np.timedelta64(365, "D"))
        if not extrapolate and distances.size != 0:
            if distances.shape[0] > 1:
                mask = np.ptp(np.sign(distances), axis = 0).astype(int) > 0
            else:
                mask = np.isfinite(y)
            dist = np.where(mask, dist, np.timedelta64(365, "D"))
        return dist
    out = np.apply_along_axis(_dtf, axis, y, x, extrapolate = extrapolate)
    return out

def cve0(lmbdas, Y, A):
    def _cve(lmbdas, Y, A):
        assert len(Y.shape) == 1
        m = len(Y)
        mask = np.isfinite(Y)
        u = np.ones_like(Y)
        Y_ = np.where(np.eye(m), np.nan, np.repeat(Y[np.newaxis, :], m, axis = 0))
        z = wt(Y_[mask,:], A, lmbdas, u, 0.5, -np.inf, np.inf, 10)
        y_hat = np.diagonal(z[:,:,mask], axis1 = 0, axis2 = 2)
        cves = np.sqrt(np.mean(((Y[mask] - y_hat)**2), axis = 1))
        return cves
    f = np.vectorize(_cve, signature = "(n),(m),(m,m)->(n)")
    cves = f(lmbdas, Y, A)
    return cves

@guvectorize([(float64[:], float64[:], float64[:,::1], float64[:], float64[:])], 
             '(k),(m),(m,m),(m)->(k)', nopython = True, target = "parallel")
def cve1(lmbdas, Y, A, u, cves):
    w = np.isfinite(Y, np.zeros_like(Y))
    W = np.diag(w*u)
    Y_ = np.where(w == 0, 0, Y)
    for i, lmbda in enumerate(lmbdas):
        H = np.linalg.solve(W + lmbda * A, W) # Eq. 13
        y_hat = np.dot(H, Y_) # Eq. 10
        hii = np.diag(H)
        cves[i,...] = np.sqrt(np.nanmean(((Y - y_hat)/(1 - hii))**2)) # Eq. 9 + 11

# if __name__ == "__main__":

#     import xarray as xr
#     from pywapor.general.processing_functions import open_ds

#     # fh = r"/Users/hmcoerver/Local/tests/test_22/SENTINEL2/S2MSI2A_R60m.nc"
#     fh = r"/Users/hmcoerver/Local/tests/test_22/VIIRSL1/VNP02IMG.nc"
#     ds = open_ds(fh)

#     y = ds.isel(y=30, x= 30).bt.values
#     x = ds.time.values

#     # Normalize x-coordinates
#     x_ = (x - x.min()) / (x.max() - x.min()) * x.size

#     # Create x-aware delta matrix.
#     A = second_order_diff_matrix(x_)

#     lmbdas = 100.
#     min_drange = -np.inf
#     max_drange = np.inf
#     # min_drange = -1.
#     # max_drange = 1.
#     max_iter = 10
#     u = np.ones(x.shape)
#     a = 0.5

#     z = wt(y, A, lmbdas, u, a, min_drange, max_drange, max_iter)

#     xdist = dist_to_finite(y, x, axis = -1, extrapolate = False)


import os
import urllib
from bs4 import BeautifulSoup
import warnings
import urllib
from joblib import Parallel, delayed
from functools import partial
import re
import numpy as np
import tqdm
from requests.exceptions import HTTPError
import time
import socket
import glob
import requests
from pywapor.general.logger import log, adjust_logger
from cachetools import cached, TTLCache
import ssl

@cached(cache=TTLCache(maxsize=2048, ttl=3600))
def find_paths(url, regex, node_type = "a", tag = "href", filter = None, session = None):
    """Search for node an tag values on a webpage.

    Parameters
    ----------
    url : str
        URL of page to search.
    regex : str
        Regex to filter data.
    node_type : str, optional
        Node type to search for, by default "a".
    tag : str, optional
        Tage within `node` to search for, by default "href".
    filter : function, optional
        Function applied to each tag found, by default None.
    session : requests.Session, optional
        session to use to open `url`, by default None.

    Returns
    -------
    np.ndarray
        Array with all tags found.
    """
    if isinstance(session, type(None)):
        context = ssl._create_unverified_context()
        f = urllib.request.urlopen(url, context = context)
    else:
        file_object = session.get(url, stream = True)
        file_object.raise_for_status()
        f = file_object.content
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", message="It looks like you're parsing an XML document using an HTML parser.")
        soup = BeautifulSoup(f, "lxml")
    file_tags = soup.find_all(node_type, {tag: re.compile(regex)})
    if not isinstance(filter, type(None)):
        file_tags = np.unique([filter(tag) for tag in file_tags], axis = 0).tolist()
    return file_tags

def crawl(urls, regex, filter_regex, session, label_filter = None, list_out = False):
    """Search for tags on multiple pages.

    Parameters
    ----------
    urls : dict
        Values are urls to search.
    regex : str
        Regex to filter data.
    session : requests.Session
        session to use to open `url`.

    Returns
    -------
    dict
        URLs found.
    """

    def _filter(tag):
        url = tag["href"]
        date_str = re.findall(filter_regex, url)[0].replace("/", "")
        return (date_str, url)

    if list_out:
        crawled_urls = list()
    else:
        crawled_urls = dict()

    for url in tqdm.tqdm(urls.values(), delay = 20):
        raw_paths = find_paths(url, regex, filter = _filter, session = session)
        if list_out:
            crawled_urls += raw_paths
        else:
            crawled_urls = {**crawled_urls, **{dt: new_url for dt, new_url in raw_paths if dt in label_filter}}

    return crawled_urls

def _download_urls(urls, folder, session, fps = None, parallel = 0, headers = None,
                    max_tries = 10, wait_sec = 15):

    if isinstance(parallel, bool):
        parallel = {True: -1, False: 0}[parallel]

    if isinstance(fps, type(None)):
        fps = [os.path.join(folder, os.path.split(url)[-1]) for url in urls]
    else:
        assert len(fps) == len(urls)

    if parallel:
        backend = "loky"
        dler = partial(download_url, session = session, headers = headers, max_tries = max_tries, wait_sec = wait_sec)
        files = Parallel(n_jobs=parallel, backend = backend)(delayed(dler)(*x) for x in zip(urls, fps))
    else:
        files = list()
        for url, fp in zip(urls, fps):
            try:
                fp_out = download_url(url, fp, session = session, headers = headers, max_tries = max_tries, wait_sec = wait_sec)
                files.append(fp_out)
            except NameError as e:
                log.info(f"--> {e}")

    return files

def unzip(zipped):
    new, l = [], []
    for x in zipped:
        new.append(x[0])
        l.append(x[1])
    return new, l

def update_urls(urls, dled_files, relevant_fps, checker_function = None):
        
    if isinstance(urls, zip):
        urls, l = unzip(urls)
        fps_given = True
    else:
        l = [os.path.split(x)[-1] for x in urls]
        fps_given = False

    for fp in dled_files:

        if fp in relevant_fps:
            continue

        if isinstance(checker_function, type(None)):
            is_relevant = True
        else:
            is_relevant = checker_function(fp)

        if fps_given:
            x = fp
        else:
            x = os.path.split(fp)[-1]
        url_idx = l.index(x) if x in l else None

        if os.path.isfile(fp):
            if not isinstance(url_idx, type(None)):
                _ = urls.pop(url_idx)
                _ = l.pop(url_idx)
            if is_relevant:
                relevant_fps.append(fp)

    if fps_given:
        urls = zip(urls, l)

    return urls, relevant_fps

def download_urls(urls, folder, session = None, fps = None, parallel = 0, 
                    headers = None, checker_function = None, 
                    max_tries = 10, wait_sec = 15, meta_max_tries = 2):
    """Download URLs to a folder.

    Parameters
    ----------
    urls : list
        URLs to download.
    folder : str
        Path to folder in which to download.
    session : request.Session, optional
        Session to use, by default None.
    fps : list, optional
        Filepaths to store each URL, should be same length as `urls`, ignores `folder`, by default None.
    parallel : int, optional
        Download URLs in parallel (currently not implemented), by default 0.
    headers : dict, optional
        Headers to include with requests, by default None.
    checker_function : function, optional
        Function to apply to downloaded files to see if they are valid, by default None.
    max_tries : int, optional
        Max number of tries to download a single file (in case of server side errors), by default 10.
    wait_sec : int, optional
        How many seconds to wait between retries, by default 15.
    meta_max_tries : int, optional
        Max number of tries to download all URLs before giving up, by default 2.

    Returns
    -------
    list
        Paths to downloaded files.
    """
    
    relevant_fps = list()
    try_n = 0
    try_again = True

    while try_again:

        dled_files = _download_urls(urls, folder, session, fps = fps, parallel = parallel, headers = headers, max_tries = max_tries, wait_sec = wait_sec)

        if not isinstance(fps, type(None)):
            urls = zip(urls, fps)

        urls, relevant_fps = update_urls(urls, dled_files, relevant_fps, checker_function = checker_function)

        if not isinstance(fps, type(None)):
            urls, fps = unzip(urls)

        try_again = len(urls) > 0 and try_n < meta_max_tries-1
        if try_again:
            try_n += 1
            log.info(f"--> Retrying to download {len(urls)} remaining urls.")

    if len(urls) > 0:
        log.warning(f"--> Didn't succeed to download {len(urls)} files.")
    return relevant_fps


def _download_url(url, fp, session = None, waitbar = True, headers = None):

    if os.path.isfile(fp):
        log.add().info(f"--> Reusing existing `{fp}` file.")
        return fp

    folder, fn = os.path.split(fp)
    if not os.path.exists(folder):
        os.makedirs(folder)

    ext = os.path.splitext(fp)[-1]
    temp_fp = fp.replace(ext, "_temp")

    if isinstance(session, type(None)):
        session = requests.Session()

    vrfy = {"NO": False, "YES": True}.get(os.environ.get("PYWAPOR_VERIFY_SSL", "YES"), True)
    file_object = session.get(url, stream = True, headers = headers, verify = vrfy)
    file_object.raise_for_status()

    if "Content-Length" in file_object.headers.keys():
        tot_size = int(file_object.headers["Content-Length"])
    else:
        tot_size = None

    if waitbar:
        wb = tqdm.tqdm(unit='Bytes', unit_scale=True, leave = False, 
                        total = tot_size, desc = fn)

    with open(temp_fp, 'wb') as z:
        for data in file_object.iter_content(chunk_size=1024):
            size = z.write(data)
            if waitbar:
                wb.update(size)

    os.rename(temp_fp, fp)

    return fp

def download_url(url, fp, session = None, waitbar = True, headers = None, 
                    max_tries = 3, wait_sec = 15):
    """Download a URL to a file.

    Parameters
    ----------
    url : str
        URL to download.
    fp : str
        Path to file to download into.
    session : requests.Session, optional
        Session to use, by default None.
    waitbar : bool, optional
        Display a download progress bar, by default True.
    headers : dict, optional
        Headers to include with requests, by default None.
    max_tries : int, optional
        Max number of tries to download a single file (in case of server side errors), by default 10.
    wait_sec : int, optional
        How many seconds to wait between retries, by default 15.

    Returns
    -------
    str
        Path to downloaded file.

    Raises
    ------
    NameError
        Raised when `max_tries` is exceeeded.
    """
    try_n = 0
    succes = False
    while try_n <= max_tries-1 and not succes:
        if try_n > 0:
            waiter = int((try_n**1.2) * wait_sec)
            log.info(f"--> Trying to download {fp}, attempt {try_n+1} of {max_tries} in {waiter} seconds.")
            time.sleep(waiter)
        try:
            fp = _download_url(url, fp, session = session, waitbar = waitbar, headers = headers)
            succes = True
        except socket.timeout as e:
            log.info(f"--> Server connection timed out.")
        except HTTPError as e:
            error_body = getattr(getattr(e, "response", ""), "text", "")
            log.info(f"--> Server error {e} [{error_body}].")
        else:
            ...
        finally:
            try_n += 1
    
    if not succes:
        raise NameError(f"Could not download {url} after {max_tries} attempts.")

    return fp

if __name__ == "__main__":

    ...



# -*- coding: utf-8 -*-
"""
This WAPOR toolbox is a set of functions to collect and run the WAPOR ET model
"""

from pywapor import general, et_look_v2_v3, pre_et_look, et_look, collect, post_et_look, pre_se_root, se_root, enhancers
from pywapor.main import Project

__all__ = ['general', 'et_look_v2_v3', 'pre_et_look', 'et_look', 'collect', 'post_et_look', 'pre_se_root', 'se_root', 'enhancers', "Project"]

__version__ = '3.5.7'



import glob
import os
import copy
import warnings
import xarray as xr
import numpy as np
import pywapor.collect.protocol.copernicus_odata as copernicus_odata
import numpy as np
from datetime import datetime as dt
from pywapor.general.logger import log
from functools import partial
from pywapor.enhancers.gap_fill import gap_fill
from pywapor.general.processing_functions import open_ds, adjust_timelim_dtype
from pywapor.general.logger import adjust_logger
from lxml import etree


def apply_qa(ds, var):
    """Mask SENTINEL2 data using a qa variable.

    Parameters
    ----------
    ds : xr.Dataset
        Input data, should contain `"qa"` and `var` as variables.
    var : str
        Variable name in `ds` to be masked.

    Returns
    -------
    xr.Dataset
        Masked dataset.
    """
    # 0 SC_NODATA # 1 SC_SATURATED_DEFECTIVE # 2 SC_DARK_FEATURE_SHADOW
    # 3 SC_CLOUD_SHADOW # 4 SC_VEGETATION # 5 SC_NOT_VEGETATED
    # 6 SC_WATER # 7 SC_UNCLASSIFIED # 8 SC_CLOUD_MEDIUM_PROBA
    # 9 SC_CLOUD_HIGH_PROBA # 10 SC_THIN_CIRRUS # 11 SC_SNOW_ICE
    if "qa" in ds.data_vars:
        pixel_qa_flags = [0, 1, 2, 3, 7, 8, 9, 10, 11]
        keep = np.invert(ds["qa"].isin(pixel_qa_flags))
        ds[var] = ds[var].where(keep)
    else:
        log.warning(f"--> Couldn't apply qa, since `qa` doesn't exist in this dataset ({list(ds.data_vars)}).")
    return ds

def mask_invalid(ds, var, valid_range = (1, 65534)):
    """Mask invalid data.

    Parameters
    ----------
    ds : xr.Dataset
        Input data.
    var : str
        Variable to mask.
    valid_range : tuple, optional
        Range of valid values in `var`, by default (1, 65534).

    Returns
    -------
    xr.Dataset
        Masked data.
    """
    # 0 = NODATA, 65535 = SATURATED
    ds[var] = ds[var].where((ds[var] >= valid_range[0]) & (ds[var] <= valid_range[1]))
    return ds

def scale_data(ds, var):
    """Apply a scale and offset factor to `ds`.

    Parameters
    ----------
    ds : xr.Dataset
        Input dataset.
    var : str
        Variable to scale and offset.

    Returns
    -------
    xr.Dataset
        Output data.
    """
    
    scale = 1./ds.scale_factor # BOA_QUANTIFICATION_VALUE
    offset = ds.offset_factor # BOA_ADD_OFFSET
    ds[var] = (ds[var] + offset) * scale
    ds[var] = ds[var].where((ds[var] <= 1.00) & (ds[var] >= 0.00))
    return ds

def calc_normalized_difference(ds, var, bands = ["nir", "red"]):
    """Calculate the normalized difference of two bands.

    Parameters
    ----------
    ds : xr.Dataset
        Input data.
    var : str
        Name of the variable in which to store the normalized difference.
    bands : list, optional
        The two bands to use to calculate the norm. difference, by default ["nir", "red"].

    Returns
    -------
    xr.Dataset
        Output data.
    """
    if np.all([x in ds.data_vars for x in bands]):
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", message="All-NaN slice encountered")
            da = (ds[bands[0]] - ds[bands[1]]) / (ds[bands[0]] + ds[bands[1]])
        ds[var] = da.clip(-1, 1)
    else:
        log.warning(f"--> Couldn't calculate `{var}`, `{'` and `'.join([x for x in bands if x not in ds.data_vars])}` is missing.")
    return ds

def calc_psri(ds, var):
    """Calculate the PSRI as ("red" - "blue)/"red_edge_740".

    Parameters
    ----------
    ds : xr.Dataset
        Input data.
    var : str
        Name of the variable in which to store the normalized difference.

    Returns
    -------
    xr.Dataset
        Output data.
    """
    reqs = ["red", "blue", "red_edge_740"]
    if np.all([x in ds.data_vars for x in reqs]):
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", message="All-NaN slice encountered")
            da = (ds["red"] - ds["blue"]) / ds["red_edge_740"]
        ds[var] = da.clip(-1, 1)
    else:
        log.warning(f"--> Couldn't calculate `{var}`, `{'` and `'.join([x for x in reqs if x not in ds.data_vars])}` is missing.")
    return ds

def calc_nmdi(ds, var):
    """Calculate the NMDI.

    Parameters
    ----------
    ds : xr.Dataset
        Input data.
    var : str
        Name of the variable in which to store the normalized difference.

    Returns
    -------
    xr.Dataset
        Output data.
    """
    reqs = ["swir1", "swir2", "nir"]
    if np.all([x in ds.data_vars for x in reqs]):
        ds["nominator"] = ds["swir1"] - ds["swir2"]
        ds = calc_normalized_difference(ds, var, bands = ["nominator", "nir"])
        ds = ds.drop_vars(["nominator"])
    else:
        log.warning(f"--> Couldn't calculate `{var}`, `{'` and `'.join([x for x in reqs if x not in ds.data_vars])}` is missing.")
    return ds

def calc_bsi(ds, var):
    """Calculate the BSI.

    Parameters
    ----------
    ds : xr.Dataset
        Input data.
    var : str
        Name of the variable in which to store the normalized difference.

    Returns
    -------
    xr.Dataset
        Output data.
    """
    reqs = ["nir", "swir1", "red", "blue"]
    if np.all([x in ds.data_vars for x in reqs]):
        ds["nominator"] = ds["nir"] + ds["blue"]
        ds["denominator"] = ds["swir1"] + ds["red"]
        ds = calc_normalized_difference(ds, var, bands = ["nominator", "denominator"])
        ds = ds.drop_vars(["nominator", "denominator"])
    else:
        log.warning(f"--> Couldn't calculate `{var}`, `{'` and `'.join([x for x in reqs if x not in ds.data_vars])}` is missing.")
    return ds

def calc_vari_red_egde(ds, var):
    """Calculate the VARI_RED_EDGE.

    Parameters
    ----------
    ds : xr.Dataset
        Input data.
    var : str
        Name of the variable in which to store the normalized difference.

    Returns
    -------
    xr.Dataset
        Output data.
    """
    reqs = ["red_edge_740", "blue", "red"]
    if np.all([x in ds.data_vars for x in reqs]):
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", message="All-NaN slice encountered")
            n1 = ds["red_edge_740"] - 1.7 * ds["red"] + 0.7 * ds["blue"]
            n2 = ds["red_edge_740"] + 2.3 * ds["red"] - 1.3 * ds["blue"]
            da = n1 / n2
        ds[var] = da.clip(-1, 1)
    else:
        log.warning(f"--> Couldn't calculate `{var}`, `{'` and `'.join([x for x in reqs if x not in ds.data_vars])}` is missing.")
    return ds

def calc_r0(ds, var):
    """Calculate the Albedo.

    Parameters
    ----------
    ds : xr.Dataset
        Input data.
    var : str
        Name of the variable in which to store the normalized difference.

    Returns
    -------
    xr.Dataset
        Output data.
    """

    weights = {
        "blue":     0.171,
        "green":    0.060,
        "red":      0.334,
        "nir":      0.371,
        "offset":   0.018,
    }
    
    reqs = ["blue", "green", "red", "nir"]
    if np.all([x in ds.data_vars for x in reqs]):
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", message="All-NaN slice encountered")
            ds["offset"] = xr.ones_like(ds["blue"])
            weights_da = xr.DataArray(data = list(weights.values()), 
                                    coords = {"band": list(weights.keys())})
            ds["r0"] = ds[reqs + ["offset"]].to_array("band").weighted(weights_da).sum("band", skipna = False)
    else:
        log.warning(f"--> Couldn't calculate `{var}`, `{'` and `'.join([x for x in reqs if x not in ds.data_vars])}` is missing.")
    return ds

def default_vars(product_name, req_vars):
    """Given a `product_name` and a list of requested variables, returns a dictionary
    with metadata on which exact layers need to be requested from the server, how they should
    be renamed, and how their dimensions are defined.

    Parameters
    ----------
    product_name : str
        Name of the product.
    req_vars : list
        List of variables to be collected.

    Returns
    -------
    dict
        Metadata on which exact layers need to be requested from the server.
    """

    variables = {

        "S2MSI2A_R20m": {
                    "_B01_20m.jp2": [(), "coastal_aerosol", [mask_invalid, apply_qa, scale_data]],
                    "_B02_20m.jp2": [(), "blue", [mask_invalid, apply_qa, scale_data]],
                    "_B03_20m.jp2": [(), "green", [mask_invalid, apply_qa, scale_data]],
                    "_B04_20m.jp2": [(), "red", [mask_invalid, apply_qa, scale_data]],
                    "_B05_20m.jp2": [(), "red_edge_703", [mask_invalid, apply_qa, scale_data]],
                    "_B06_20m.jp2": [(), "red_edge_740", [mask_invalid, apply_qa, scale_data]],
                    "_B07_20m.jp2": [(), "red_edge_782", [mask_invalid, apply_qa, scale_data]],
                    "_B8A_20m.jp2": [(), "nir", [mask_invalid, apply_qa, scale_data]],
                    "_B11_20m.jp2": [(), "swir1", [mask_invalid, apply_qa, scale_data]],
                    "_B12_20m.jp2": [(), "swir2", [mask_invalid, apply_qa, scale_data]],
                    "_SCL_20m.jp2": [(), "qa", []],
                },
        "S2MSI2A_R60m": {
                    "_B01_60m.jp2": [(), "coastal_aerosol", [mask_invalid, apply_qa, scale_data]],
                    "_B02_60m.jp2": [(), "blue", [mask_invalid, apply_qa, scale_data]],
                    "_B03_60m.jp2": [(), "green", [mask_invalid, apply_qa, scale_data]],
                    "_B04_60m.jp2": [(), "red", [mask_invalid, apply_qa, scale_data]],
                    "_B05_60m.jp2": [(), "red_edge_703", [mask_invalid, apply_qa, scale_data]],
                    "_B06_60m.jp2": [(), "red_edge_740", [mask_invalid, apply_qa, scale_data]],
                    "_B07_60m.jp2": [(), "red_edge_782", [mask_invalid, apply_qa, scale_data]],
                    "_B8A_60m.jp2": [(), "nir", [mask_invalid, apply_qa, scale_data]],
                    "_B09_60m.jp2": [(), "narrow_nir", [mask_invalid, apply_qa, scale_data]],
                    "_B11_60m.jp2": [(), "swir1", [mask_invalid, apply_qa, scale_data]],
                    "_B12_60m.jp2": [(), "swir2", [mask_invalid, apply_qa, scale_data]],
                    "_SCL_60m.jp2": [(), "qa", []],
                },
    }

    req_dl_vars = {

        "S2MSI2A_R20m": {
            "coastal_aerosol":  ["_B01_20m.jp2", "_SCL_20m.jp2"],
            "blue":             ["_B02_20m.jp2", "_SCL_20m.jp2"],
            "green":            ["_B03_20m.jp2", "_SCL_20m.jp2"],
            "red":              ["_B04_20m.jp2", "_SCL_20m.jp2"],
            "red_edge_703":     ["_B05_20m.jp2", "_SCL_20m.jp2"],
            "red_edge_740":     ["_B06_20m.jp2", "_SCL_20m.jp2"],
            "red_edge_782":     ["_B07_20m.jp2", "_SCL_20m.jp2"],
            "nir":              ["_B8A_20m.jp2", "_SCL_20m.jp2"],
            "swir1":            ["_B11_20m.jp2", "_SCL_20m.jp2"],
            "swir2":            ["_B12_20m.jp2", "_SCL_20m.jp2"],
            "qa":               ["_SCL_20m.jp2"],
            "ndvi":             ["_B04_20m.jp2", "_B8A_20m.jp2", "_SCL_20m.jp2"],
            "mndwi":            ["_B03_20m.jp2", "_B11_20m.jp2", "_SCL_20m.jp2"],
            "vari_red_edge":    ["_B06_20m.jp2", "_B02_20m.jp2", "_B04_20m.jp2", "_SCL_20m.jp2"],
            "nmdi":             ["_B11_20m.jp2", "_B12_20m.jp2", "_B8A_20m.jp2", "_SCL_20m.jp2"],
            "psri":             ["_B02_20m.jp2", "_B04_20m.jp2", "_B06_20m.jp2", "_SCL_20m.jp2"],
            "bsi":              ["_B8A_20m.jp2", "_B11_20m.jp2", "_B04_20m.jp2", "_B02_20m.jp2", "_SCL_20m.jp2"],
            "r0":               ["_B02_20m.jp2", "_B03_20m.jp2", "_B04_20m.jp2", "_B8A_20m.jp2", "_SCL_20m.jp2"],
        },

        "S2MSI2A_R60m": {
            "coastal_aerosol":  ["_B01_60m.jp2", "_SCL_60m.jp2"],
            "blue":             ["_B02_60m.jp2", "_SCL_60m.jp2"],
            "green":            ["_B03_60m.jp2", "_SCL_60m.jp2"],
            "red":              ["_B04_60m.jp2", "_SCL_60m.jp2"],
            "red_edge_703":     ["_B05_60m.jp2", "_SCL_60m.jp2"],
            "red_edge_740":     ["_B06_60m.jp2", "_SCL_60m.jp2"],
            "red_edge_782":     ["_B07_60m.jp2", "_SCL_60m.jp2"],
            "nir":              ["_B8A_60m.jp2", "_SCL_60m.jp2"],
            "swir1":            ["_B11_60m.jp2", "_SCL_60m.jp2"],
            "swir2":            ["_B12_60m.jp2", "_SCL_60m.jp2"],
            "qa":               ["_SCL_60m.jp2"],
            "ndvi":             ["_B04_60m.jp2", "_B8A_60m.jp2", "_SCL_60m.jp2"],
            "mndwi":            ["_B03_60m.jp2", "_B11_60m.jp2", "_SCL_60m.jp2"],
            "vari_red_edge":    ["_B06_60m.jp2", "_B02_60m.jp2", "_B04_60m.jp2", "_SCL_60m.jp2"],
            "nmdi":             ["_B11_60m.jp2", "_B12_60m.jp2", "_B8A_60m.jp2", "_SCL_60m.jp2"],
            "psri":             ["_B02_60m.jp2", "_B04_60m.jp2", "_B06_60m.jp2", "_SCL_60m.jp2"],
            "bsi":              ["_B8A_60m.jp2", "_B11_60m.jp2", "_B04_60m.jp2", "_B02_60m.jp2", "_SCL_60m.jp2"],
            "r0":               ["_B02_60m.jp2", "_B03_60m.jp2", "_B04_60m.jp2", "_B8A_60m.jp2", "_SCL_60m.jp2"],
        },
    }

    out = {val:variables[product_name][val] for sublist in map(req_dl_vars[product_name].get, req_vars) for val in sublist}

    return out

def default_post_processors(product_name, req_vars):
    """Given a `product_name` and a list of requested variables, returns a dictionary with a 
    list of functions per variable that should be applied after having collected the data
    from a server.

    Parameters
    ----------
    product_name : str
        Name of the product.
    req_vars : list
        List of variables to be collected.

    Returns
    -------
    dict
        Functions per variable that should be applied to the variable.
    """
    
    post_processors = {
        "S2MSI2A_R20m": {
            "coastal_aerosol":  [gap_fill],
            "blue":             [gap_fill],
            "green":            [gap_fill],
            "red":              [gap_fill],
            "red_edge_703":     [gap_fill],
            "red_edge_740":     [gap_fill],
            "red_edge_782":     [gap_fill],
            "nir":              [gap_fill],
            "qa":               [],
            "swir1":            [gap_fill],
            "swir2":            [gap_fill],
            "psri":             [calc_psri],
            "ndvi":             [calc_normalized_difference, gap_fill],
            "nmdi":             [calc_nmdi, gap_fill],
            "vari_red_edge":    [calc_vari_red_egde, gap_fill],
            "bsi":              [calc_bsi, gap_fill],
            "mndwi":            [partial(calc_normalized_difference, bands = ["swir1", "green"]), gap_fill],
            "r0":               [calc_r0, gap_fill],
            },

        "S2MSI2A_R60m": {
            "coastal_aerosol":  [gap_fill],
            "blue":             [gap_fill],
            "green":            [gap_fill],
            "red":              [gap_fill],
            "red_edge_703":     [gap_fill],
            "red_edge_740":     [gap_fill],
            "red_edge_782":     [gap_fill],
            "nir":              [gap_fill],
            "qa":               [],
            "swir1":            [gap_fill],
            "swir2":            [gap_fill],
            "psri":             [calc_psri, gap_fill],
            "ndvi":             [calc_normalized_difference, gap_fill],
            "nmdi":             [calc_nmdi, gap_fill],
            "vari_red_edge":    [calc_vari_red_egde, gap_fill],
            "bsi":              [calc_bsi, gap_fill],
            "mndwi":            [partial(calc_normalized_difference, bands = ["swir1", "green"]), gap_fill],
            "r0":               [calc_r0, gap_fill],
            },
    }

    out = {k:v for k,v in post_processors[product_name].items() if k in req_vars}

    return out

def time_func(fn):
    """Return a np.datetime64 given a filename.

    Parameters
    ----------
    fn : str
        Filename.

    Returns
    -------
    np.datetime64
        Date as described in the filename.
    """
    dtime = np.datetime64(dt.strptime(fn.split("_")[2], "%Y%m%dT%H%M%S"), "ns")
    return dtime

def s2_processor(scene_folder, variables, **kwargs):

    dss = list()
    dss__ = list()
    for k, v in variables.items():
        ds__ = open_ds(glob.glob(os.path.join(scene_folder, "**", "*" + k), recursive = True)[0], decode_coords=None)
        ds_ = ds__.isel(band=0).rename({"band_data": v[1]})
        dss__.append(ds__)
        dss.append(ds_)

    ds = xr.merge(dss).drop_vars("band")

    meta_fps = glob.glob(os.path.join(scene_folder, "**", "MTD_MSIL2A.xml"), recursive = True)
    
    if len(meta_fps) >= 1:
        tree = etree.parse(meta_fps[0])
        root = tree.getroot()
        baseline = [float(x.text) for x in root.iter('PROCESSING_BASELINE')][0]    
        # NOTE https://sentinels.copernicus.eu/en/web/sentinel/-/copernicus-sentinel-2-major-products-upgrade-upcoming
        # NOTE https://stackoverflow.com/questions/72566760/how-to-correct-sentinel-2-baseline-v0400-offset
        if baseline >= 4.0:
            offsets = [float(x.text) for x in root.iter('BOA_ADD_OFFSET')]
            scales = [float(x.text) for x in root.iter('BOA_QUANTIFICATION_VALUE')]
            scale = np.median(scales)
            offset = np.median(offsets)
            ds.attrs = {"scale_factor": scale, "offset_factor": offset}
        else:
            offsets = [0.]
            offset = offsets[0]
            scales = [float(x.text) for x in root.iter('BOA_QUANTIFICATION_VALUE')]
            scale = np.median(scales)
            ds.attrs = {"scale_factor": scale, "offset_factor": offset}
        if len(np.unique(offsets)) != 1:
            log.warning(f"--> Multiple offsets found for `{scene_folder}`, using `{offset}`, check `{meta_fps[0]}` for more info.")
        if len(scales) != 1:
            log.warning(f"--> Multiple scales found for `{scene_folder}`, using `{scale}`, check `{meta_fps[0]}` for more info.")
    else:
        t = time_func(os.path.split(scene_folder)[-1])
        if t < np.datetime64("2022-01-25"):
            ds.attrs = {"scale_factor": 10000.0, "offset_factor": 0.}
        else:
            ds.attrs = {"scale_factor": 10000.0, "offset_factor": -1000.0}
        log.warning(f"--> No scale/offset found for `{meta_fps[0]}`, using `{ds.attrs}`.")

    return ds, dss__

def download(folder, latlim, lonlim, timelim, product_name, 
                req_vars, variables = None, post_processors = None, 
                extra_search_kwargs = {"cloudcoverpercentage": (0, 30)}):
    """Download SENTINEL2 data and store it in a single netCDF file.

    Parameters
    ----------
    folder : str
        Path to folder in which to store results.
    latlim : list
        Latitude limits of area of interest.
    lonlim : list
        Longitude limits of area of interest.
    timelim : list
        Period for which to prepare data.
    product_name : str
        Name of the product to download.
    req_vars : list
        Which variables to download for the selected product.
    variables : dict, optional
        Metadata on which exact layers need to be requested from the server, by default None.
    post_processors : dict, optional
        Functions per variable that should be applied to the variable, by default None.
    extra_search_kwargs : dict
        Extra search kwargs passed to SentinelAPI, by default {"cloudcoverpercentage": (0, 30)}.

    Returns
    -------
    xr.Dataset
        Downloaded data.
    """

    product_folder = os.path.join(folder, "SENTINEL2")

    fn = os.path.join(product_folder, f"{product_name}.nc")
    req_vars_orig = copy.deepcopy(req_vars)
    if os.path.isfile(fn):
        existing_ds = open_ds(fn)
        req_vars_new = list(set(req_vars).difference(set(existing_ds.data_vars)))
        if len(req_vars_new) > 0:
            req_vars = req_vars_new
            existing_ds = existing_ds.close()
        else:
            return existing_ds[req_vars_orig]

    if isinstance(variables, type(None)):
        variables = default_vars(product_name, req_vars)

    if isinstance(post_processors, type(None)):
        post_processors = default_post_processors(product_name, req_vars)
    else:
        default_processors = default_post_processors(product_name, req_vars)
        post_processors = {k: {True: default_processors[k], False: v}[v == "default"] for k,v in post_processors.items() if k in req_vars}

    timelim = adjust_timelim_dtype(timelim)
    bb = [lonlim[0], latlim[0], lonlim[1], latlim[1]]

    def node_filter(node_info):
        fn = os.path.split(node_info)[-1]
        to_dl = list(variables.keys()) + ["MTD_MSIL2A.xml"]
        return np.any([x in fn for x in to_dl])

    final_fn = os.path.split(fn)[-1]
    processor = s2_processor
    scenes = copernicus_odata.download(product_folder, latlim, lonlim, timelim, "SENTINEL2", product_name.split("_")[0], node_filter = node_filter)
    ds = copernicus_odata.process_sentinel(scenes, variables, time_func, final_fn, post_processors, processor, bb = bb)

    return ds[req_vars_orig]

if __name__ == "__main__":

    folder = r"/Users/hmcoerver/Local/s2_test"
    adjust_logger(True, folder, "INFO")
    timelim = ["2022-03-29", "2022-03-31"]
    latlim = [28.9, 29.7]
    lonlim = [30.2, 31.2]

    product_name = 'S2MSI2A_R60m'
    req_vars = ["mndwi", "psri", "vari_red_edge", "bsi", "nmdi", "green", "nir"]
    post_processors = None
    variables = None



from functools import partial
import functools
from pywapor.general.logger import log
from pywapor.general.processing_functions import func_from_string

def apply_enhancers(post_processors, ds):
    for var, funcs in post_processors.items():
        for func in funcs:
            if isinstance(func, str):
                func = func_from_string(func)
            elif isinstance(func, dict):
                args = func.get("args", ())
                kwargs = func.get("keywords", {})
                func = func_from_string(func["func"])
                if kwargs or args:
                    func = functools.partial(func, *args, **kwargs)
            ds, label = apply_enhancer(ds, var, func)
            log.info(label)
    return ds

def apply_enhancer(ds, variable, enhancer):
    """Apple a function to a (variable in a) dataset. 

    Parameters
    ----------
    ds : xr.Dataset
        Dataset to enhance.
    variable : str
        Variable name to enhance.
    enhancer : function
        Function to be applied to `variable` inside `ds`. Should take `ds` as
        first argument and `variable` as second.

    Returns
    -------
    tuple
        The enhanced dataset and the label to log when calculating the dataset.
    """
    ds = enhancer(ds, variable)

    if isinstance(enhancer, partial):
        func_name = enhancer.func.__name__
    else:
        func_name = enhancer.__name__
    
    if isinstance(variable, type(None)):
        label = f"--> Applying '{func_name}'."
    else:
        label = f"--> Applying '{func_name}' to `{variable}`."

    return ds, label



import numpy as np
from pywapor.et_look_v2_v3 import constants as con
import xarray as xr
import warnings

def longitude_rad(lon_deg):
    r"""
    Converts longitude from degrees to radians.

    Parameters
    ----------
    lon_deg : float
        longitude in degrees, 
        :math:`\phi`
        [deg]

    Returns
    -------
    lon : float
        longitude, 
        :math:`\phi`
        [rad]

    """
    return lon_deg * np.pi/180.0


def latitude_rad(lat_deg):
    r"""
    Converts latitude from degrees to radians.

    Parameters
    ----------
    lat_deg : float
        latitude in degrees, 
        :math:`\lambda`
        [deg]

    Returns
    -------
    lat : float
        latitude, 
        :math:`\lambda`
        [rad]

    """
    return lat_deg * np.pi/180.0


def slope_rad(slope_deg):
    r"""
    Converts slope from degrees to radians.

    Parameters
    ----------
    slope_deg : float
        slope in degrees, 
        :math:`s`
        [deg]

    Returns
    -------
    slope : float
        slope, 
        :math:`\Delta`
        [rad]

    """
    return slope_deg * np.pi/180.0


def aspect_rad(aspect_deg):
    r"""
    Converts aspect from degrees to radians.

    Parameters
    ----------
    aspect_deg : float
        aspect in degrees, 
        :math:`s`
        [deg]

    Returns
    -------
    aspect : float
        aspect (0 is north; pi is south), 
        :math:`\alpha`
        [rad]
    """
    return aspect_deg * np.pi/180.0


def declination(doy):
    r"""
    Computes the solar declination which is the angular height of the sun
    above the astronomical equatorial plane in radians.

    .. math ::

        \delta=0.409 \cdot \sin\left(\frac{2\pi \cdot J}{365}-1.39\right)

    Parameters
    ----------
    doy : float
        julian day of the year, 
        :math:`J`
        [-]

    Returns
    -------
    decl : float
        declination, 
        :math:`\delta`
        [rad]

    Examples
    --------
    >>> import ETLook.solar_radiation as solrad
    >>> solrad.declination(180)
    0.40512512455439242
    """
    return 0.409 * np.sin(2 * np.pi / 365 * doy - 1.39)


def earth_sun_distance(doy):
    r"""
    Computes the earth sun distance in Angstrom Unit where 1 AU is 1.496e8 km.

    .. math ::

        d_{r}=1+0.033 \cdot \cos\left(\frac{2\pi \cdot J}{365}\right)

    Parameters
    ----------
    doy : float
        julian day of the year, 
        :math:`J`
        [-]

    Returns
    -------
    ed : float
        earth sun distance, 
        :math:`d_{r}`
        [AU]

    Examples
    --------
    >>> import ETLook.solar_radiation as solrad
    >>> solrad.earth_sun_distance(180)
    0.96703055420162642
    """

    warnings.warn(('Using "ed" for the inverse earth-sun distance is deprecated, '
                   'use "iesd" (solar_radiation.inverse_earth_sun_distance) instead'),
                  DeprecationWarning, stacklevel=2)

    return 1 + 0.033 * np.cos(doy * 2 * np.pi / 365.0)


def inverse_earth_sun_distance(doy):
    r"""
    Computes the inverse earth sun distance (iesd) in Angstrom Unit where 1 AU is 1.496e8 km.

    .. math ::

        d_{r}=1+0.033 \cdot \cos\left(\frac{2\pi \cdot J}{365}\right)

    Parameters
    ----------
    doy : float
        julian day of the year, 
        :math:`J`
        [-]

    Returns
    -------
    iesd : float
        inverse earth sun distance, 
        :math:`d_{r}`
        [AU]

    Examples
    --------
    >>> import ETLook.solar_radiation as solrad
    >>> solrad.inverse_earth_sun_distance(180)
    0.96703055420162642
    """

    return 1 + 0.033 * np.cos(doy * 2 * np.pi / 365.0)


def actual_earth_sun_distance(iesd):
    r"""
    Computes the earth sun distance (esd) in Angstrom Unit where 1 AU is 1.496e8 km.

    .. math ::

        d_{r}=1+0.033 \cdot \cos\left(\frac{2\pi \cdot J}{365}\right)

    Parameters
    ----------
    iesd : float
        inverse earth sun distance, 
        :math:`J`
        [AU]

    Returns
    -------
    esd : float
        earth sun distance, 
        :math:`d_{r}`
        [AU]

    Examples
    --------
    >>> import ETLook.solar_radiation as solrad
    >>> solrad.actual_earth_sun_distance(180)
    1.034093489244084
    """

    return 1 / iesd


def seasonal_correction(doy):
    r"""
    Computes the seasonal correction for solar time  in hours.

    .. math ::

        b=\frac{2\pi\left(J-81\right)}{365}

    .. math ::

        s_{c}= 0.1645 \cdot sin \left( 2b \right) - 0.1255 \cdot cos \left(b \right) - 0.025 \left(b\right)

    Parameters
    ----------
    doy : float
        julian day of the year, 
        :math:`J`
        [-]

    Returns
    -------
    sc : float
        seasonal correction, 
        :math:`s_{c}`
        [hours]

    Examples
    --------
    >>> import ETLook.solar_radiation as solrad
    >>> solrad.seasonal_correction(180)
    -0.052343379605521212
    """
    b = 2 * np.pi * (doy - 81) / 365.0
    return 0.1645 * np.sin(2 * b) - 0.1255 * np.cos(b) - 0.025 * np.sin(b)


def sunset_hour_angle(lat, decl):
    r"""
    Computes the sunset hour angle.

    .. math ::

        w_{s}=\arccos(-\tan(\lambda)\cdot \tan(\delta))

    Parameters
    ----------
    decl : float
        solar declination, 
        :math:`\delta`
        [rad]
    lat : float
        latitude, 
        :math:`\lambda`
        [rad]

    Returns
    -------
    ws : float
        sunset hour angle, 
        :math:`w_{s}`
        [rad]

    """
    return np.arccos(-(np.tan(lat) * np.tan(decl)))


def hour_angle(sc, dtime, lon = 0):
    r"""
    Computes the hour angle which is zero at noon and -pi at 0:00 am and
    pi at 12:00 pm

    .. math ::

        \omega=\left(\frac{\pi}{12}\right)\cdot \left(t+s_{c}-12\right)

    Parameters
    ----------
    sc : float
        seasonal correction, 
        :math:`s_{c}`
        [hours]
    dtime : float
        decimal time, 
        :math:`t`
        [hours]
    lon : float
        longitude, 
        :math:`\phi`
        [rad]

    Returns
    -------
    ha : float
        hour_angle, 
        :math:`\omega`
        [rad]

    Examples
    --------
    >>> import ETLook.solar_radiation as solrad
    >>> solrad.hour_angle(sc=solrad.seasonal_correction(75), dtime=11.4)
    -0.19793970172084141
    """
    dtime = dtime + (lon / (15*np.pi/180.0))
    return (np.pi / 12.0) * (dtime + sc - 12.0)


def inst_solar_radiation_toa(csza, iesd):
    r"""
    Computes the instantaneous solar radiation at the top of
    the atmosphere [Wm-2].

    .. math ::

        S_{toa}^{i} = S_{sun} \cdot d_{r} \cdot \phi

    Parameters
    ----------
    csza : float
        cosine solar zenith angle, 
        :math:`\phi`
        [-]
    iesd : float
        inverse earth sun distance, 
        :math:`d_{r}`
        [AU]

    Returns
    -------
    ra_i_toa : float
        instantaneous solar radiation at top of atmosphere, 
        :math:`S_{toa}^{i}`
        [Wm-2]

    Examples
    --------
    >>> import ETLook.solar_radiation as solrad
    >>> doy = 1
    >>> sc = solrad.seasonal_correction(doy)
    >>> ha = solrad.hour_angle(sc, dtime=12)
    >>> decl = solrad.declination(doy)
    >>> csza = solrad.cosine_solar_zenith_angle(ha, decl, 0)
    >>> iesd = solrad.inverse_earth_sun_distance(doy)
    >>> solrad.inst_solar_radiation_toa(csza, iesd)
    1299.9181944414036
    """
    return csza * con.sol * iesd


def daily_solar_radiation_toa(sc, decl, iesd, lat, slope=0, aspect=0):
    r"""
    Computes the daily solar radiation at the top of the atmosphere.

    .. math ::

        S_{toa}=S_{sun} \cdot d_{r}\int_{i=-\pi}^{i=\pi}S_{toa}^{i}

    Parameters
    ----------
    iesd : float
        inverse earth sun distance, 
        :math:`d_{r}`
        [AU]
    decl : float
        solar declination, 
        :math:`\delta`
        [rad]
    sc : float
        seasonal correction, 
        :math:`s_{c}`
        [hours]
    lat : float
        latitude, 
        :math:`\lambda`
        [rad]
    slope : float
        slope, 
        :math:`\Delta`
        [rad]
    aspect : float
        aspect (0 is north; pi is south), 
        :math:`\alpha`
        [rad]

    Returns
    -------
    ra_24_toa : float
        daily solar radiation at the top of atmosphere, 
        :math:`S_{toa}`
        [Wm-2]

    Examples
    --------
    >>> import ETLook.solar_radiation as solrad
    >>> from math import pi
    >>> doy = 1
    >>> sc = solrad.seasonal_correction(doy)
    >>> decl = solrad.declination(doy)
    >>> iesd = solrad.inverse_earth_sun_distance(doy)
    >>> solrad.daily_solar_radiation_toa(sc, decl, iesd, lat=25*pi/180.0)
    265.74072308978026
    """

    # hour angle for the whole day in half-hourly intervals (0:15-23:45)
    t_start = 0.25
    t_end = 24.00
    interval = 0.5
    times = [t_start+i*interval for i in range(0, 48)]
    hours = [hour_angle(sc, t) for t in times]

    ra24 = 0

    for t in hours:
        csza = cosine_solar_zenith_angle(t, decl, lat, slope, aspect)
        ra24 += inst_solar_radiation_toa(csza, iesd) / len(hours)

    # return the average daily radiation
    return ra24


def cosine_solar_zenith_angle(ha, decl, lat, slope=0, aspect=0):
    r"""
    computes the cosine of the solar zenith angle [-].

    .. math ::

        \phi = & \sin\left(\delta\right) \cdot \sin\left(\lambda\right) \cdot  \cos\left(\Delta\right) - \\
        & \sin\left(\delta\right) \cdot \cos\left(\lambda\right) \cdot \sin\left(\Delta\right) + \\
        & \cos\left(\delta\right) \cdot \cos\left(\lambda\right) \cdot \cos\left(\Delta\right) \cdot \cos\left(\omega\right)+\\
        & \cos\left(\delta\right) \cdot \sin\left(\lambda\right) \cdot \sin\left(\Delta\right) \cdot \sin\left(\alpha\right) \cdot \cos\left(\omega\right)+\\
        & \cos\left(\delta\right) \cdot \sin\left(\Delta\right) \cdot  \sin\left(\alpha\right) \cdot \sin\left(\omega\right)

    Parameters
    ----------
    ha : float
        hour angle, 
        :math:`\omega`
        [rad]
    decl : float
        declination, 
        :math:`\delta`
        [rad]
    lat : float
        latitude, 
        :math:`\lambda`
        [rad]
    slope : float
        slope, 
        :math:`\Delta`
        [rad]
    aspect : float
        aspect (0 is north; pi is south), 
        :math:`\alpha`
        [rad]

    Returns
    -------
    csza : float
        cosine solar zenith angle, 
        :math:`\phi`
        [-]

    Examples
    --------
    >>> import ETLook.solar_radiation as solrad
    >>> sc = solrad.seasonal_correction(1)
    >>> ha = solrad.hour_angle(sc, dtime=12)
    >>> solrad.cosine_solar_zenith_angle(ha, decl=solrad.declination(1), lat=0)
    0.92055394167363314
    """
    t1 = np.sin(decl) * np.sin(lat) * np.cos(slope)
    t2 = np.sin(decl) * np.cos(lat) * np.sin(slope) * np.cos(aspect - np.pi)
    t3 = np.cos(decl) * np.cos(lat) * np.cos(slope)
    t4 = np.cos(decl) * np.sin(lat) * np.sin(slope) * np.cos(aspect - np.pi)
    t5 = np.cos(decl) * np.sin(slope) * np.sin(aspect - np.pi)
    csza = t1 - t2 + t3 * np.cos(ha) + t4 * np.cos(ha) + t5 * np.sin(ha)

    # check if the sun is above the horizon
    check = np.sin(decl) * np.sin(lat) + np.cos(decl) * \
        np.cos(lat) * np.cos(ha)

    if isinstance(csza, xr.DataArray):
        res = xr.where((csza > 0) & (check >= 0), csza, 0)
        res = xr.where(csza.notnull() | check.notnull(), res, np.nan)
    else:
        nans = np.logical_or(np.isnan(csza), np.isnan(check))
        res = np.where(np.logical_and(csza > 0, check >= 0), csza, 0)
        res[nans] = np.nan

    return res


def transmissivity(ra, ra_flat):
    """Computes the transmissivity.

    Parameters
    ----------
    ra_24 : float
        daily solar radiation for a flat surface, 
        :math:`S^{\downarrow}`
        [Wm-2]
    ra_24_toa_flat : float
        daily solar radiation at the top of atmosphere for a flat surface, 
        :math:`S_{toa,f}`
        [Wm-2]

    Returns
    -------
    trans_24 : float
        daily atmospheric transmissivity, 
        :math:`\tau`
        [-]
    """
    return ra / ra_flat


def daily_solar_radiation_toa_flat(decl, iesd, lat, ws):
    r"""
    Computes the daily solar radiation at the top of the atmosphere for a flat
    surface.

    .. math ::

        S_{toa,f}=\frac{S_{sun}}{\pi} \cdot d_{inv,r} \cdot (w_{s} \cdot \sin(\lambda) \cdot \sin(\delta) +
                  \cos(\lambda)\cdot\cos(\delta)\cdot\sin(w_{s}))

    Parameters
    ----------
    decl : float
        solar declination, 
        :math:`\delta`
        [rad]
    iesd : float
        inverse earth sun distance, 
        :math:`d_{inv,r}`
        [AU]
    lat : float
        latitude, 
        :math:`\lambda`
        [rad]
    ws : float
        sunset hour angle, 
        :math:`w_{s}`
        [rad]

    Returns
    -------
    ra_24_toa_flat : float
        daily solar radiation at the top of atmosphere for a flat surface, 
        :math:`S_{toa,f}`
        [Wm-2]

    """
    ra_flat = (con.sol / np.pi) * iesd * (ws * np.sin(lat) * np.sin(decl) +
                                        np.cos(lat) * np.cos(decl) * np.sin(ws))

    return ra_flat


def daily_solar_radiation_flat(ra_24_toa_flat, trans_24):
    r"""
    Computes the daily solar radiation at the earth's surface.

    .. math ::

        S^{\downarrow} = \tau \cdot S_{toa}

    Parameters
    ----------
    ra_24_toa_flat : float
        daily solar radiation at the top of atmosphere for a flat surface, 
        :math:`S_{toa}`
        [Wm-2]
    trans_24 : float
        daily atmospheric transmissivity, 
        :math:`\tau`
        [-]

    Returns
    -------
    ra_24 : float
        daily solar radiation for a flat surface, 
        :math:`S^{\downarrow}`
        [Wm-2]

    """
    return ra_24_toa_flat * trans_24


def diffusion_index(trans_24, diffusion_slope=-1.33, diffusion_intercept=1.15):
    r"""
    Computes the diffusion index, the ratio between diffuse and direct
    solar radiation. The results are clipped between 0 and 1.

    .. math ::

        I_{diff} = a_{diff}+b_{diff} \cdot \tau

    Parameters
    ----------
    trans_24 : float
        daily atmospheric transmissivity, 
        :math:`\tau`
        [-]
    diffusion_slope : float
        slope of diffusion index vs transmissivity relationship, 
        :math:`b_{diff}`
        [-]
    diffusion_intercept : float
        intercept of diffusion index vs transmissivity relationship, 
        :math:`a_{diff}`
        [-]

    Returns
    -------
    diffusion_index : float
        diffusion_index, 
        :math:`I_{diff}`
        [-]

    """
    res = diffusion_intercept + trans_24 * diffusion_slope

    res = np.clip(res, 0, 1)

    return res


def daily_total_solar_radiation(ra_24_toa, ra_24_toa_flat, diffusion_index, trans_24):
    r"""
    Computes the daily solar radiation at the earth's surface taken
    diffuse and direct solar radiation into account.

    .. math ::

        S^{\downarrow} = I_{diff} \cdot \tau \cdot S_{toa,f} +(1-I_{diff}) \cdot \tau \cdot S_{toa}

    Parameters
    ----------
    ra_24_toa : float
        daily solar radiation at the top of atmosphere, 
        :math:`S_{toa}`
        [Wm-2]
    ra_24_toa_flat : float
        daily solar radiation at the top of atmosphere for a flat surface, 
        :math:`S_{toa,f}`
        [Wm-2]
    diffusion_index : float
        diffusion_index, 
        :math:`I_{diff}`
        [-]
    trans_24 : float
        daily atmospheric transmissivity, 
        :math:`\tau`
        [-]

    Returns
    -------
    ra_24 : float
        daily solar radiation, 
        :math:`S^{\downarrow}`
        [Wm-2]

    """
    diffuse = trans_24 * ra_24_toa_flat * diffusion_index
    direct = trans_24 * ra_24_toa * (1 - diffusion_index)
    return diffuse + direct




"""
@author: Radoslaw Guzinski
Copyright: (C) 2017, Radoslaw Guzinski
"""
import math
import os
import numpy as np
from osgeo import gdal
from sklearn import tree, linear_model, ensemble
import pywapor.enhancers.dms.pyDMS_utils as utils
from pywapor.general.logger import log
import warnings

REG_sknn_ann = 0
REG_sklearn_ann = 1

class DecisionTreeRegressorWithLinearLeafRegression(tree.DecisionTreeRegressor):
    ''' Decision tree regressor with added linear (bayesian ridge) regression
    for all the data points falling within each decision tree leaf node.

    Parameters
    ----------
    linearRegressionExtrapolationRatio: float (optional, default: 0.25)
        A limit on extrapolation allowed in the per-leaf linear regressions.
        The ratio is multiplied by the range of values present in each leaves'
        training dataset and added (substracted) to the maxiumum (minimum)
        value.

    decisionTreeRegressorOpt: dictionary (optional, default: {})
        Options to pass to DecisionTreeRegressor constructor. See
        http://scikit-learn.org/stable/modules/generated/sklearn.tree.DecisionTreeRegressor.html
        for possibilities.

    Returns
    -------
    None
    '''
    def __init__(self, linearRegressionExtrapolationRatio=0.25, decisionTreeRegressorOpt={}):
        super(DecisionTreeRegressorWithLinearLeafRegression, self).__init__(**decisionTreeRegressorOpt)
        self.decisionTreeRegressorOpt = decisionTreeRegressorOpt
        self.leafParameters = {}
        self.linearRegressionExtrapolationRatio = linearRegressionExtrapolationRatio

    def fit(self, X, y, sample_weight, fitOpt={}):
        ''' Build a decision tree regressor from the training set (X, y).

        Parameters
        ----------
        X: array-like or sparse matrix, shape = [n_samples, n_features]
            The training input samples. Internally, it will be converted to
            dtype=np.float32 and if a sparse matrix is provided to a sparse
            csc_matrix.

        y: array-like, shape = [n_samples] or [n_samples, n_outputs]
            The target values (real numbers). Use dtype=np.float64 and
            order='C' for maximum efficiency.

        sample_weight: array-like, shape = [n_samples] or None
            Sample weights. If None, then samples are equally weighted. Splits
            that would create child nodes with net zero or negative weight are
            ignored while searching for a split in each node.

        fitOpt: dictionary (optional, default: {})
            Options to pass to DecisionTreeRegressor fit function. See
            http://scikit-learn.org/stable/modules/generated/sklearn.tree.DecisionTreeRegressor.html
            for possibilities.

        Returns
        -------
        Self
        '''

        # Fit a normal regression tree
        super(DecisionTreeRegressorWithLinearLeafRegression, self).fit(X, y, sample_weight,
                                                                       **fitOpt)

        # Create a linear regression for all input points which fall into
        # one output leaf
        predictedValues = super(DecisionTreeRegressorWithLinearLeafRegression, self).predict(X)
        leafValues = np.unique(predictedValues)
        for value in leafValues:
            ind = predictedValues == value
            leafLinearRegrsion = linear_model.BayesianRidge()
            leafLinearRegrsion.fit(X[ind, :], y[ind])
            self.leafParameters[value] = {"linearRegression": leafLinearRegrsion,
                                          "max": np.max(y[ind]),
                                          "min": np.min(y[ind])}

        return self

    def predict(self, X, predictOpt={}):
        ''' Predict class or regression value for X.

        Parameters
        ----------
        X: array-like or sparse matrix of shape = [n_samples, n_features]
            The input samples. Internally, it will be converted to
            dtype=np.float32 and if a sparse matrix is provided to a sparse
            csr_matrix.

        predictOpt: dictionary (optional, default: {})
            Options to pass to DecisionTreeRegressor predict function. See
            http://scikit-learn.org/stable/modules/generated/sklearn.tree.DecisionTreeRegressor.html
            for possibilities.

        Returns
        -------
        y: array of shape = [n_samples] or [n_samples, n_outputs]
            The predicted classes, or the predict values.
        '''

        # Do normal regression tree prediction
        y = super(DecisionTreeRegressorWithLinearLeafRegression, self).predict(X, **predictOpt)

        # And also apply per-leaf linear regression
        for leafValue in self.leafParameters.keys():
            ind = y == leafValue
            if X[ind, :].size > 0:
                y[ind] = self.leafParameters[leafValue]["linearRegression"].predict(X[ind, :])
                # Limit extrapolation
                extrapolationRange = self.linearRegressionExtrapolationRatio * (
                                        self.leafParameters[leafValue]["max"] -
                                        self.leafParameters[leafValue]["min"])
                y[ind] = np.maximum(y[ind],
                                    self.leafParameters[leafValue]["min"] - extrapolationRange)
                y[ind] = np.minimum(y[ind],
                                    self.leafParameters[leafValue]["max"] + extrapolationRange)

        return y


class DecisionTreeSharpener(object):
    ''' Decision tree based sharpening (disaggregation) of low-resolution
    images using high-resolution images. The implementation is mostly based on :footcite:t:`gao2012`, 2012.

    Decision tree based regressor is trained with high-resolution data resampled to
    low resolution and low-resolution data and then applied
    directly to high-resolution data to obtain high-resolution representation
    of the low-resolution data.

    The implementation includes selecting training data based on homogeneity
    statistics and using the homogeneity as weight factor (:footcite:t:`gao2012`, 2012, section 2.2),
    performing linear regression with samples located within each regression
    tree leaf node (:footcite:t:`gao2012`, 2012, section 2.1), using an ensemble of regression trees
    (:footcite:t:`gao2012`, 2012, section 2.1), performing local (moving window) and global regression and
    combining them based on residuals (:footcite:t:`gao2012`, 2012, section 2.3) and performing residual
    analysis and bias correction (:footcite:t:`gao2012`, 2012, section 2.4)


    Parameters
    ----------
    highResFiles: list of strings
        A list of file paths to high-resolution images to be used during the
        training of the sharpener.

    lowResFiles: list of strings
        A list of file paths to low-resolution images to be used during the
        training of the sharpener. There must be one low-resolution image
        for each high-resolution image.

    lowResQualityFiles: list of strings (optional, default: [])
        A list of file paths to low-resolution quality images to be used to
        mask out low-quality low-resolution pixels during training. If provided
        there must be one quality image for each low-resolution image.

    lowResGoodQualityFlags: list of integers (optional, default: [])
        A list of values indicating which pixel values in the low-resolution
        quality images should be considered as good quality.

    cvHomogeneityThreshold: float (optional, default: 0)
        A threshold of coeficient of variation below which high-resolution
        pixels resampled to low-resolution are considered homogeneous and
        usable during the training of the disaggregator. If threshold is 0 or
        negative then it is set automatically such that 80% of pixels are below
        it.

    movingWindowSize: integer (optional, default: 0)
        The size of local regression moving window in low-resolution pixels. If
        set to 0 then only global regression is performed.

    disaggregatingTemperature: boolean (optional, default: False)
        Flag indicating whether the parameter to be disaggregated is
        temperature (e.g. land surface temperature). If that is the case then
        at some points it needs to be converted into radiance. This is becasue
        sensors measure energy, not temperature, plus radiance is the physical
        measurements it makes sense to average, while radiometric temperature
        behaviour is not linear.

    perLeafLinearRegression: boolean (optional, default: True)
        Flag indicating if linear regression should be performed on all data
        points falling within each regression tree leaf node.

    linearRegressionExtrapolationRatio: float (optional, default: 0.25)
        A limit on extrapolation allowed in the per-leaf linear regressions.
        The ratio is multiplied by the range of values present in each leaves'
        training dataset and added (substracted) to the maxiumum (minimum)
        value.

    regressorOpt: dictionary (optional, default: {})
        Options to pass to DecisionTreeRegressor constructor See
        http://scikit-learn.org/stable/modules/generated/sklearn.tree.DecisionTreeRegressor.html
        for possibilities. Note that max_leaf_nodes and min_samples_leaf
        parameters will beoverwritten in the code.

    baggingRegressorOpt: dictionary (optional, default: {})
        Options to pass to BaggingRegressor constructor. See
        http://scikit-learn.org/stable/modules/generated/sklearn.ensemble.BaggingRegressor.html
        for possibilities.

    Returns
    -------
    None

    '''
    def __init__(self,
                 highResFiles,
                 lowResFiles,
                 lowResQualityFiles=[],
                 lowResGoodQualityFlags=[],
                 cvHomogeneityThreshold=0,
                 movingWindowSize=0,
                 disaggregatingTemperature=False,
                 perLeafLinearRegression=True,
                 linearRegressionExtrapolationRatio=0.25,
                 regressorOpt={},
                 baggingRegressorOpt={}):

        self.highResFiles = highResFiles
        self.lowResFiles = lowResFiles
        self.lowResQualityFiles = lowResQualityFiles
        self.lowResGoodQualityFlags = lowResGoodQualityFlags

        if len(self.highResFiles) != len(self.lowResFiles):
            raise IOError("There must be a matching high resolution file for each low resolution file")

        if len(self.lowResQualityFiles) == 0 or \
           (len(self.lowResQualityFiles) == 1 and self.lowResQualityFiles[0] == ""):
            self.useQuality_LR = False
        else:
            self.useQuality_LR = True

        if self.useQuality_LR and len(self.lowResQualityFiles) != len(self.lowResFiles):
            raise IOError("The number of quality files must be 0 or the same as number of low " +
                  "resolution files")

        self.cvHomogeneityThreshold = cvHomogeneityThreshold
        # If threshold is 0 or negative then it is set automatically such that
        # 80% of pixels are below it.
        if self.cvHomogeneityThreshold <= 0:
            self.autoAdjustCvThreshold = True
            self.precentileThreshold = 80
        else:
            self.autoAdjustCvThreshold = False

        # Moving window size in low resolution pixels
        self.movingWindowSize = float(movingWindowSize)
        # The extension (on each side) by which sampling window size is larger
        # then prediction window size (see section 2.3 of Gao paper)
        self.movingWindowExtension = self.movingWindowSize * 0.25
        self.windowExtents = []

        self.disaggregatingTemperature = disaggregatingTemperature

        # Flag to determine whether a multivariate linear regression should be
        # constructed for samples in each leaf of the regression tree
        # (see section 2.1 of Gao paper)
        self.perLeafLinearRegression = perLeafLinearRegression
        self.linearRegressionExtrapolationRatio = linearRegressionExtrapolationRatio

        self.regressorOpt = regressorOpt
        self.baggingRegressorOpt = baggingRegressorOpt

    def trainSharpener(self):
        ''' Train the sharpener using high- and low-resolution input files
        and settings specified in the constructor. Local (moving window) and
        global regression decision trees are trained with high-resolution data
        resampled to low resolution and low-resolution data. The training
        dataset is selected based on homogeneity of resampled high-resolution
        data being below specified threshold and quality mask (if given) of
        low resolution data. The homogeneity statistics are also used as weight
        factors for the training samples (more homogenous - higher weight).

        Parameters
        ----------
        None

        Returns
        -------
        None
        '''

        # Select good data (training samples) from low- and high-resolution
        # input images.
        fileNum = 0
        for highResFile, lowResFile in zip(self.highResFiles, self.lowResFiles):

            scene_HR = gdal.Open(highResFile)
            scene_LR = gdal.Open(lowResFile)

            # First subset and reproject low res scene to fit with
            # high res scene
            subsetScene_LR = utils.reprojectSubsetLowResScene(scene_HR, scene_LR)
            data_LR = subsetScene_LR.GetRasterBand(1).ReadAsArray()
            gt_LR = subsetScene_LR.GetGeoTransform()

            # Do the same with low res quality file (if provided) and flag
            # pixels which are considered to be of good quality
            if self.useQuality_LR:
                quality_LR = gdal.Open(self.lowResQualityFiles[fileNum])
                subsetQuality_LR = utils.reprojectSubsetLowResScene(scene_HR, quality_LR)
                subsetQualityMask = subsetQuality_LR.GetRasterBand(1).ReadAsArray()
                qualityPix = np.in1d(subsetQualityMask.ravel(),
                                     self.lowResGoodQualityFlags).reshape(subsetQualityMask.shape)
                quality_LR = None
            else:
                qualityPix = np.ones(data_LR.shape).astype(bool)

            # Low resolution pixels with NaN value or no-data-value are always of bad quality
            no_data = subsetScene_LR.GetRasterBand(1).GetNoDataValue()
            qualityPix = np.all([qualityPix, ~np.isnan(data_LR), data_LR != no_data], axis = 0)

            scale_factor = scene_LR.GetRasterBand(1).GetScale()
            if not isinstance(scale_factor, type(None)) and scale_factor != 0:
                data_LR = data_LR.astype(float) * scale_factor

            # Then resample high res scene to low res pixel size while
            # extracting sub-low-res-pixel homogeneity statistics
            resMean, resStd = utils.resampleHighResToLowRes(scene_HR, subsetScene_LR)
            resMean[resMean == 0] = 0.000001
            resCV = np.sum(resStd/resMean, 2) / resMean.shape[2]
            resCV[np.isnan(resCV)] = 1000

            # Resampled high resolution pixels where at least one "parameter"
            # is NaN are also of bad quality
            resNaN = np.any(np.isnan(resMean), -1)
            qualityPix = np.logical_and(qualityPix, ~resNaN)

            windows = []
            extents = []
            # If moving window approach is used (section 2.3 of Gao paper)
            # then calculate the extent of each sampling window in low
            # resolution pixels
            if self.movingWindowSize > 0:
                for y in range(int(math.ceil(data_LR.shape[0]/self.movingWindowSize))):
                    for x in range(int(math.ceil(data_LR.shape[1]/self.movingWindowSize))):
                        windows.append([int(max(y*self.movingWindowSize-self.movingWindowExtension, 0)),
                                        int(min((y+1)*self.movingWindowSize+self.movingWindowExtension,
                                                data_LR.shape[0])),
                                        int(max(x*self.movingWindowSize-self.movingWindowExtension, 0)),
                                        int(min((x+1)*self.movingWindowSize+self.movingWindowExtension,
                                                data_LR.shape[1]))])
                        # Save the extents of this window in projection coordinates as
                        # UL and LR point coordinates
                        ul = utils.pix2point([x*self.movingWindowSize, y*self.movingWindowSize],
                                             gt_LR)
                        lr = utils.pix2point([(x+1)*self.movingWindowSize,
                                              (y+1)*self.movingWindowSize],
                                             gt_LR)
                        extents.append([ul, lr])

            # And always add the whole extent of low res image to also estimate
            # the regression tree for the whole image
            windows.append([0, data_LR.shape[0], 0, data_LR.shape[1]])

            goodData_LR = [None for _ in range(len(windows))]
            goodData_HR = [None for _ in range(len(windows))]
            weight = [None for _ in range(len(windows))]

            # For each window extract the good quality low res and high res pixels
            for i, window in enumerate(windows):
                rows = slice(window[0], window[1])
                cols = slice(window[2], window[3])
                qualityPixWindow = qualityPix[rows, cols]
                resCVWindow = resCV[rows, cols]

                # Good pixels are those where low res data quality is good and
                # high res data is homonogenous
                if self.autoAdjustCvThreshold:
                    g = np.logical_and.reduce((qualityPixWindow, resCVWindow < 1000,
                                               resCVWindow > 0))
                    if ~np.any(g):
                        self.cvHomogeneityThreshold = 0
                    else:
                        self.cvHomogeneityThreshold = np.percentile(resCVWindow[g],
                                                                    self.precentileThreshold)
                    log.info(f'--> Homogeneity CV threshold: {self.cvHomogeneityThreshold:.2f}')
                homogenousPix = np.logical_and(resCVWindow < self.cvHomogeneityThreshold,
                                               resCVWindow > 0)
                goodPix = np.logical_and(homogenousPix, qualityPixWindow)

                goodData_LR[i] = utils.appendNpArray(goodData_LR[i],
                                                     data_LR[rows, cols][goodPix])
                goodData_HR[i] = utils.appendNpArray(goodData_HR[i],
                                                     resMean[rows, cols, :][goodPix, :], axis=0)

                # Also estimate weight given to each pixel as the inverse of its
                # heterogeneity
                w = 1/resCVWindow[goodPix]
                weight[i] = utils.appendNpArray(weight[i], w)

                # Print some stats
                if np.prod(data_LR[rows, cols][qualityPixWindow].shape) > 0:
                    percentageUsedPixels = int(float(np.prod(goodData_LR[i].shape)) /
                                               float(np.prod(data_LR[rows, cols][qualityPixWindow].shape)) * 100)
                    log.info(f'--> Number of training elements for window {i} is ' +
                          str(np.prod(goodData_LR[i].shape)) + ', representing ' +
                          str(percentageUsedPixels)+'% of avaiable low-resolution data.')

            # Close all files
            scene_HR = None
            scene_LR = None
            subsetScene_LR = None
            if self.useQuality_LR:
                subsetQuality_LR = None
            fileNum = fileNum + 1

        self.windowExtents = extents
        windowsNum = len(windows)

        # Once all the samples have been picked fit all the local and global
        # regressions
        self.reg = [None for _ in range(windowsNum)]
        for i in range(windowsNum):
            if i < windowsNum-1:
                local = True
            else:
                local = False
            if len(goodData_LR[i]) > 0:
                self.reg[i] = \
                    self._doFit(goodData_LR[i], goodData_HR[i], weight[i], local)

    def applySharpener(self, highResFilename, lowResFilename=None):
        ''' Apply the trained sharpener to a given high-resolution image to
        derive corresponding disaggregated low-resolution image. If local
        regressions were used during training then they will only be applied
        where their moving window extent overlaps with the high resolution
        image passed to this function. Global regression will be applied to the
        whole high-resolution image wihtout geographic constraints.

        Parameters
        ----------
        highResFilename: string
            Path to the high-resolution image file do be used during
            disaggregation.

        lowResFilename: string (optional, default: None)
            Path to the low-resolution image file corresponding to the
            high-resolution input file. If local regressions
            were trained and low-resolution filename is given then the local
            and global regressions will be combined based on residual values of
            the different regressions to the low-resolution image (see :footcite:t:`gao2012`, 2012,
            section 2.3). If local regressions were trained and low-resolution
            filename is not given then only the local regressions will be used.


        Returns
        -------
        outImage: GDAL memory file object
            The file object contains an in-memory, georeferenced disaggregator
            output.
        '''

        # Open and read the high resolution input file
        highResFile = gdal.Open(highResFilename)
        inData = np.zeros((highResFile.RasterYSize, highResFile.RasterXSize,
                           highResFile.RasterCount))
        for band in range(highResFile.RasterCount):
            data = highResFile.GetRasterBand(band+1).ReadAsArray().astype(float)
            no_data = highResFile.GetRasterBand(band+1).GetNoDataValue()
            data[data == no_data] = np.nan
            inData[:, :, band] = data
        gt = highResFile.GetGeoTransform()

        shape = inData.shape
        ysize = shape[0]
        xsize = shape[1]

        # Temporarly get rid of NaN's
        nanInd = np.isnan(inData)
        inData[nanInd] = 0
        outWindowData = np.full((ysize, xsize), np.nan)

        # Do the downscailing on the moving windows if there are any
        for i, extent in enumerate(self.windowExtents):
            if self.reg[i] is not None:
                [minX, minY] = utils.point2pix(extent[0], gt)  # UL
                [minX, minY] = [max(minX, 0), max(minY, 0)]
                [maxX, maxY] = utils.point2pix(extent[1], gt)  # LR
                [maxX, maxY] = [min(maxX, xsize), min(maxY, ysize)]
                windowInData = inData[minY:maxY, minX:maxX, :]
                outWindowData[minY:maxY, minX:maxX] = \
                    self._doPredict(windowInData, self.reg[i])

        # Do the downscailing on the whole input image
        if self.reg[-1] is not None:
            outFullData = self._doPredict(inData, self.reg[-1])
        else:
            outFullData =  np.full((ysize, xsize), np.nan)

        # Combine the windowed and whole image regressions
        # If there is no windowed regression just use the whole image regression
        if np.all(np.isnan(outWindowData)):
            outData = outFullData
        # If corresponding low resolution file is provided then combine the two
        # regressions based on residuals (see section 2.3 of Gao paper)
        elif lowResFilename is not None:
            lowResScene = gdal.Open(lowResFilename)
            outWindowScene = utils.saveImg(outWindowData,
                                           highResFile.GetGeoTransform(),
                                           highResFile.GetProjection(),
                                           "MEM",
                                           noDataValue=np.nan)
            windowedResidual, _, _ = self._calculateResidual(outWindowScene, lowResScene)
            outWindowScene = None
            outFullScene = utils.saveImg(outFullData,
                                         highResFile.GetGeoTransform(),
                                         highResFile.GetProjection(),
                                         "MEM",
                                         noDataValue=np.nan)
            fullResidual, _, _ = self._calculateResidual(outFullScene, lowResScene)
            outFullScene = None
            lowResScene = None
            # windowed weight
            ww = (1/windowedResidual)**2/((1/windowedResidual)**2 + (1/fullResidual)**2)
            # full weight
            fw = 1 - ww
            outData = outWindowData*ww + outFullData*fw
        # Otherwised use just windowed regression
        else:
            outData = outWindowData

        # Fix NaN's
        nanInd = np.any(nanInd, -1)
        outData[nanInd] = np.nan

        outImage = utils.saveImg(outData,
                                 highResFile.GetGeoTransform(),
                                 highResFile.GetProjection(),
                                 "MEM",
                                 noDataValue=np.nan)

        highResFile = None
        inData = None
        return outImage

    def residualAnalysis(self, disaggregatedFile, lowResFilename, lowResQualityFilename=None,
                         doCorrection=True):
        ''' Perform residual analysis and (optional) correction on the
        disaggregated file (see :footcite:t:`gao2012`, 2012, section 2.4).

        Parameters
        ----------
        disaggregatedFile: string or GDAL file object
            If string, path to the disaggregated image file; if gdal file
            object, the disaggregated image.

        lowResFilename: string
            Path to the low-resolution image file corresponding to the
            high-resolution disaggregated image.

        lowResQualityFilename: string (optional, default: None)
            Path to low-resolution quality image file. If provided then low
            quality values are masked out during residual analysis. Otherwise
            all values are considered to be of good quality.

        doCorrection: boolean (optional, default: True)
            Flag indication whether residual (bias) correction should be
            performed or not.


        Returns
        -------
        residualImage: GDAL memory file object
            The file object contains an in-memory, georeferenced residual image.

        correctedImage: GDAL memory file object
            The file object contains an in-memory, georeferenced residual
            corrected disaggregated image, or None if doCorrection was set to
            False.
        '''

        if not os.path.isfile(str(disaggregatedFile)):
            scene_HR = disaggregatedFile
        else:
            scene_HR = gdal.Open(disaggregatedFile)
        scene_LR = gdal.Open(lowResFilename)

        if isinstance(lowResQualityFilename, str):
            quality_LR = gdal.Open(lowResQualityFilename)
        elif lowResQualityFilename == True:
            quality_LR = True
        else:
            quality_LR = None

        residual_HR, residual_LR, gt_res = self._calculateResidual(scene_HR, scene_LR, quality_LR)

        if self.disaggregatingTemperature:
            if doCorrection:
                with warnings.catch_warnings():
                    warnings.filterwarnings("ignore", message="invalid value encountered in power")                    
                    corrected = (residual_HR + scene_HR.GetRasterBand(1).ReadAsArray()**4)**0.25
                correctedImage = utils.saveImg(corrected,
                                               scene_HR.GetGeoTransform(),
                                               scene_HR.GetProjection(),
                                               "MEM",
                                               noDataValue=np.nan)
            else:
                correctedImage = None
            # Convert residual back to temperature for easier visualisation
            with warnings.catch_warnings():
                warnings.filterwarnings("ignore", message="invalid value encountered in power")            
                residual_LR = (residual_LR + 273.15**4)**0.25 - 273.15
        else:
            if doCorrection:
                corrected = residual_HR + scene_HR.GetRasterBand(1).ReadAsArray()
                correctedImage = utils.saveImg(corrected,
                                               scene_HR.GetGeoTransform(),
                                               scene_HR.GetProjection(),
                                               "MEM",
                                               noDataValue=np.nan)
            else:
                correctedImage = None

        residualImage = utils.saveImg(residual_LR,
                                      gt_res,
                                      scene_HR.GetProjection(),
                                      "MEM",
                                      noDataValue=np.nan)

        log.info(f"--> Low res. residual bias = {np.nanmean(residual_LR):.3f} and RMSD = {np.nanmean(residual_LR**2)**0.5:.3f}.")

        scene_HR = None
        scene_LR = None
        quality_LR = None

        return residualImage, correctedImage

    def _doFit(self, goodData_LR, goodData_HR, weight, local):
        ''' Private function. Fits the regression tree.
        '''

        # For local regression constrain the number of tree
        # nodes (rules) - section 2.3
        if local:
            self.regressorOpt["max_leaf_nodes"] = 10
        else:
            self.regressorOpt["max_leaf_nodes"] = 30
        self.regressorOpt["min_samples_leaf"] = 10

        # If per leaf linear regression is used then use modified
        # DecisionTreeRegressor. Otherwise use the standard one.
        if self.perLeafLinearRegression:
            baseRegressor = \
                DecisionTreeRegressorWithLinearLeafRegression(self.linearRegressionExtrapolationRatio,
                                                              self.regressorOpt)
        else:
            baseRegressor = \
                tree.DecisionTreeRegressor(**self.regressorOpt)

        reg = ensemble.BaggingRegressor(baseRegressor, **self.baggingRegressorOpt)
        if goodData_HR.shape[0] <= 1:
            reg.max_samples = 1.0
        reg = reg.fit(goodData_HR, goodData_LR, sample_weight=weight)

        return reg

    def _doPredict(self, inData, reg):
        ''' Private function. Calls the regression tree.
        '''

        origShape = inData.shape
        if len(origShape) == 3:
            bands = origShape[2]
        else:
            bands = 1
        # Do the actual decision tree regression
        inData = inData.reshape((-1, bands))
        outData = reg.predict(inData)
        outData = outData.reshape((origShape[0], origShape[1]))

        return outData

    def _calculateResidual(self, downscaledScene, originalScene, originalSceneQuality=None):
        ''' Private function. Calculates residual between overlapping
            high-resolution and low-resolution images.
        '''

        # First subset and reproject original (low res) scene to fit with
        # downscaled (high res) scene
        subsetScene_LR = utils.reprojectSubsetLowResScene(downscaledScene,
                                                          originalScene,
                                                          resampleAlg=gdal.GRA_NearestNeighbour)
        data_LR = subsetScene_LR.GetRasterBand(1).ReadAsArray().astype(float)
        gt_LR = subsetScene_LR.GetGeoTransform()

        # If quality file for the low res scene is provided then mask out all
        # bad quality pixels in the subsetted LR scene. Otherwise assume that all
        # low res pixels are of good quality.
        if isinstance(originalSceneQuality, gdal.Dataset):
            subsetQuality_LR = utils.reprojectSubsetLowResScene(downscaledScene,
                                                                originalSceneQuality,
                                                                resampleAlg=gdal.GRA_NearestNeighbour)
            goodPixMask_LR = subsetQuality_LR.GetRasterBand(1).ReadAsArray()
            goodPixMask_LR = np.in1d(goodPixMask_LR.ravel(),
                                     self.lowResGoodQualityFlags).reshape(goodPixMask_LR.shape)
            data_LR[~goodPixMask_LR] = np.nan
        elif originalSceneQuality == True:
            nodata = subsetScene_LR.GetRasterBand(1).GetNoDataValue()
            data_LR[data_LR == nodata] = np.nan

        scale_factor = subsetScene_LR.GetRasterBand(1).GetScale()
        if not isinstance(scale_factor, type(None)) and scale_factor != 0:
            data_LR *= scale_factor

        # Then resample high res scene to low res pixel size
        if self.disaggregatingTemperature:
            # When working with tempratures they should be converted to
            # radiance values before aggregating to be physically accurate.
            radianceScene = utils.saveImg(downscaledScene.GetRasterBand(1).ReadAsArray()**4,
                                          downscaledScene.GetGeoTransform(),
                                          downscaledScene.GetProjection(),
                                          "MEM",
                                          noDataValue=np.nan)
            resMean, _ = utils.resampleHighResToLowRes(radianceScene,
                                                       subsetScene_LR)
            # Find the residual (difference) between the two)
            residual_LR = data_LR**4 - resMean[:, :, 0]
        else:
            resMean, _ = utils.resampleHighResToLowRes(downscaledScene,
                                                       subsetScene_LR)
            # Find the residual (difference) between the two
            residual_LR = data_LR - resMean[:, :, 0]

        # Smooth the residual and resample to high resolution
        residual = utils.binomialSmoother(residual_LR)
        residualDs = utils.saveImg(residual, subsetScene_LR.GetGeoTransform(),
                                   subsetScene_LR.GetProjection(), "MEM", noDataValue=np.nan)
        residualScene_BL = utils.resampleWithGdalWarp(residualDs, downscaledScene,
                                                      resampleAlg="bilinear")
        residualDs = None

        residual = residualScene_BL.GetRasterBand(1).ReadAsArray()
        # Sometimes there can be 1 HR pixel NaN border arond LR invalid pixels due to resampling.
        # Fuction below fixes this. Image border pixels are excluded due to numba stencil
        # limitations.
        residual[1:-1, 1:-1] = utils.removeEdgeNaNs(residual)[1:-1, 1:-1]
        residualScene_BL = None

        # The residual array might be slightly smaller then the downscaled because
        # of the subsetting of the low resolution scene. In that case just pad
        # the missing values with neighbours.
        downscaled = downscaledScene.GetRasterBand(1).ReadAsArray()
        if downscaled.shape != residual.shape:
            temp = np.zeros(downscaled.shape)
            temp[:residual.shape[0], :residual.shape[1]] = residual
            temp[residual.shape[0]:, :] = \
                temp[2*(residual.shape[0] - downscaled.shape[0]):residual.shape[0] - downscaled.shape[0], :]
            temp[:, residual.shape[1]:] = \
                temp[:, 2*(residual.shape[1] - downscaled.shape[1]):residual.shape[1] - downscaled.shape[1]]

            residual = temp

        residualScene = None
        subsetScene_LR = None
        subsetQuality_LR = None

        return residual, residual_LR, gt_LR



"""
    The soil_moisture module contains all functions related to soil moisture data components,
    as outlined by :footcite:t:`yang2015estimation`.

"""
from pywapor.et_look_v2_v3 import constants as c
from pywapor.et_look_v2_v3 import unstable
import numpy as np
import xarray as xr


def wet_bulb_temperature_inst(t_air_i, t_dew_i):
    r"""
    Computes the instantaneous wet bulb temperature.

    Parameters
    ----------
    t_air_i : float
        instantaneous air temperature, 
        :math:`T_{a}`
        [C]
    t_dew_i : float
        instantaneous dew point temperature, 
        :math:`Td_{a}`
        [C]

    Returns
    -------
    t_wet_i : float
        instantaneous wet bulb temperature, 
        :math:`Tw_{a}`
        [C]
    """
    tw = wetbulb_temperature_iter(t_air_i, t_dew_i)

    return tw

def wet_bulb_temperature_inst_new(t_air_i, qv_i, p_air_i):
    r"""
    Computes the instantaneous wet bulb temperature based on Stull, Roland. "Wet-bulb temperature from relative 
    humidity and air temperature." Journal of applied meteorology and climatology 50.11 (2011): 2267-2269.

    Parameters
    ----------
    t_air_i : float
        instantaneous air temperature, 
        :math:`T_{a}`
        [C]
    qv_i : float
        instantaneous specific humidity, 
        :math:`q_{v,i}`
        [kg/kg]
    p_air_i : float
        instantaneous air pressure, 
        :math:`P_{i}`
        [mbar]

    Returns
    -------
    t_wet_i : float
        instantaneous wet bulb temperature, 
        :math:`Tw_{a}`
        [C]
    """
    es_i = 0.6108*np.exp((17.27*t_air_i)/(t_air_i+237.3))
    rh_i = np.minimum((1.6077717*qv_i/10*p_air_i/es_i),1)*100
    rh_i = rh_i.clip(0,100)
    
    tw = t_air_i * np.arctan(0.152 * (rh_i + 8.3136)**(1/2)) + np.arctan(t_air_i + rh_i) - np.arctan(rh_i - 1.6763) + 0.00391838 *(rh_i)**(3/2) * np.arctan(0.0231 * rh_i) - 4.686

    return tw

def dew_point_temperature_inst(vp_i):
    r"""
    Computes the instantaneous dew point temperature.

    Parameters
    ----------
    vp_i : float
        instantaneous vapour pressure, 
        :math:`e_{a}`
        [mbar]

    Returns
    -------
    t_dew_i : float
        instantaneous dew point temperature, 
        :math:`Td_{a}`
        [K]
    """
    t_dew_i = (237.3 * np.log(vp_i / 6.108)) / (17.27 - np.log(vp_i / 6.108))

    return t_dew_i


def dew_point_temperature_coarse_inst(vp_i):
    r"""
    Computes the instantaneous dew point temperature.

    Parameters
    ----------
    vp_i : float
        instantaneous vapour pressure, 
        :math:`e_{a}`
        [mbar]

    Returns
    -------
    t_dew_coarse_i : float
        instantaneous dew point temperature, 
        :math:`Td_{a}`
        [K]
    """
    t_dew_i = (237.3 * np.log(vp_i / 6.108)) / (17.27 - np.log(vp_i / 6.108))

    return t_dew_i


# only used internally
def latent_heat_iter(t):
    # temperature in celcius
    lv = 1000 * (2501 - 2.361 * t)

    return lv


# only used internally
def psychometric_constant_iter(lv, p=1013.25, cp=1004, rm=0.622):
    # lv: latent heat of vaporization (J/s?)
    # p:  pressure in hPa
    # cp: specific heat
    # rm: ratio of molecular weight

    psy = (cp * p) / (lv * rm)

    return psy


# only used internally
def vapor_pressure_iter(t):
    # t: temperature in Celcius

    vp = 6.108 * np.exp((17.27 * t) / (237.3 + t))

    return vp


# only used internally
def wetbulb_temperature_iter(ta, td):
    maxiter = 1000
    tol = 1e-7
    pressure = 1013.25

    lv = latent_heat_iter(ta)
    psy = psychometric_constant_iter(lv, p=pressure)

    tw = td + ((ta - td) / 3)
    ea_ta = vapor_pressure_iter(td)

    n = 0

    prev_dir = np.zeros_like(ta)
    step = (ta - td) / 5.

    while abs(np.nanmax(step)) > tol:

        ea_tw = vapor_pressure_iter(tw) - psy * (ta - tw)

        direction = (-1) ** ((ea_tw - ea_ta) > 0)

        step = np.where(prev_dir != direction, step * 0.5, step)

        tw += step * direction
        prev_dir = direction

        n += 1
        if n >= maxiter:
            return tw

    return tw


# some internal functions for the stability correction based on Brutsaert
def psi_m(y):
    r"""
    Computes the stability correction for momentum based on :footcite:t:`brutsaert1999aspects`.

    .. math ::

        \Psi_{M}(y) = & \ln(a+y)-3by^{\frac{1}{3}} + \\
        & \frac{ba^{\frac{1}{3}}}{2}\ln[\frac{(1+x)^{2}}{(1-x+x^{2})}] + \\
        & \sqrt{3}ba^{\frac{1}{3}}\arctan[\frac{(2x-1)}{\sqrt{3}}]+\Psi_{0}

    where the following constants are used

    * :math:`a` = 0.33
    * :math:`b` = 0.41

    in which:

    .. math ::

        x = (\frac{y}{a})^{\frac{1}{3}}

    and:

    .. math ::

        y = \frac{-(z-d)}{L}

    where :math:`L` is the monin obukhov length defined by
    :func:`ETLook.unstable.monin_obukhov_length`,
    :math:`z` and :math:`d` are the
    measurement height and displacement height respectively. All aforementioned
    parameters are different for the bare soil and full canopy solutions.

    The symbol :math:`\Psi_{0}` denotes a constant of integration, given by

    .. math ::

        \Psi_{0}=-\ln({a}) + \sqrt{3} \cdot b \cdot a^{\frac{1}{3}} \cdot \frac{\pi}{6}

    Notes
    -----
    This function should not be used as an input function for a ETLook tool.
    This function is used internally by :func:`aerodynamical_resistance_bare`
    and :func:`aerodynamical_resistance_full` and :func:`wind_speed_soil`.

    """
    a = 0.33
    b = 0.41
    x = (y / a) ** (1. / 3.)
    phi_0 = -np.log(a) + np.sqrt(3) * b * a ** (1. / 3.) * np.pi / 6.
    res = (
        np.log(a + y)
        - 3 * b * y ** (1. / 3.)
        + (b * a ** (1. / 3.)) / 2. * np.log((1 + x) ** 2 / (1 - x + x ** 2))
        + np.sqrt(3) * b * a ** (1. / 3.) * np.arctan((2 * x - 1) / np.sqrt(3))
        + phi_0
    )
    return res


def psi_h(y):
    r"""
    Computes the stability correction for momentum based on
    :footcite:t:`brutsaert1999aspects`.

    .. math ::

        \Psi_{H}(y)=[\frac{(1-d)}{n}]\ln{\frac{(c+y^n)}{c}}

    where the following constants are used

    * :math:`c` = 1.00
    * :math:`d` = 0.057
    * :math:`n` = 0.78

    in which

    .. math ::

        y = \frac{-(z-d)}{L}

    where :math:`L` is the monin obukhov length defined by
    :func:`ETLook.unstable.monin_obukhov_length`,
    :math:`z` and :math:`d` are the
    measurement height and displacement height respectively. All aforementioned
    parameters are different for the bare soil and full canopy solutions.

    Notes
    -----
    This function should not be used as an input function for a tool.
    This function is used internally by :func:`aerodynamical_resistance_bare`
    and :func:`aerodynamical_resistance_full` and :func:`wind_speed_soil`.

    """
    c = 0.33
    d = 0.057
    n = 0.78
    return ((1 - d) / n) * np.log((c + y ** n) / c)


def initial_friction_velocity_inst(u_b_i, z0m, disp, z_b=100):
    r"""
    Computes the initial instantaneous friction velocity without stability
    corrections.

    .. math ::

        u_{*}=\frac{k \cdot u_{b}}{\ln\left(\frac{z_{b}-d}{z_{0,m}}\right)}

    Parameters
    ----------
    u_b_i : float
        instantaneous wind speed at blending height, 
        :math:`u_{b}`
        [m s-1]
    z0m : float
        surface roughness, 
        :math:`z_{0,m}`
        [m]
    disp : float
        displacement height, 
        :math:`d`
        [m]
    z_b : float
        blending height, 
        :math:`z_{b}`
        [m]


    Returns
    -------
    u_star_i_init : float
        initial estimate of the instantaneous friction velocity, 
        :math:`u_{*,i}`
        [m s-1]
    """
    return (c.k * u_b_i) / (np.log((z_b - disp) / z0m))


def atmospheric_emissivity_inst(vp_i, t_air_k_i):
    r"""
    Computes the atmospheric emissivity according to :footcite:t:`brutsaert1975derivable`.

    .. math ::

        \varepsilon_{a}=a\left(\frac{e_{a}}{T_{a}}\right)^{b}

    where the following constants are used

    * :math:`a` = 1.24
    * :math:`b` = 1/7

    Parameters
    ----------
    vp_i : float
        instantaneous vapour pressure, 
        :math:`e_{a}`
        [mbar]
    t_air_k_i : float
        instantaneous air temperature, 
        :math:`T_{a}`
        [K]

    Returns
    -------
    emiss_atm_i : float
        instantaneous atmospheric emissivity, 
        :math:`\varepsilon_{a}`
        [-]

    """
    return 1.24 * (vp_i / t_air_k_i) ** (1. / 7.)


def net_radiation_bare(ra_hor_clear_i, emiss_atm_i, t_air_k_i, lst, r0_bare=0.38):
    r"""
    Computes the net radiation for the bare soil with zero evaporation.

    .. math ::

        Q_{bare}^{*}=\left(1-\alpha_{0,bare}\right)S_{d}+\varepsilon_{s} \cdot \varepsilon_{a} \cdot \sigma \cdot T_{a}^{4}-\varepsilon_{s} \cdot \sigma \cdot T_{s}^{4}

    Parameters
    ----------
    ra_hor_clear_i : float
        Total clear-sky irradiance on a horizontal surface, 
        :math:`S_{d}`
        [W/m2]
    emiss_atm_i : float
        instantaneous atmospheric emissivity, 
        :math:`\varepsilon_{a}`
        [-]
    t_air_k_i : float
        instantaneous air temperature, 
        :math:`T_{a}`
        [K]
    lst : float
        surface temperature, 
        :math:`T_{0}`
        [K]
    r0_bare : float
        dry bare soil surface albedo, 
        :math:`\alpha_{0, bare}`
        [-]


    Returns
    -------
    rn_bare : float
        net radiation bare soil, 
        :math:`Q^*_{bare}`
        [Wm-2]
    """
    emiss_bare = 0.95
    rn_bare = (
        (1 - r0_bare) * ra_hor_clear_i
        + emiss_atm_i * emiss_bare * c.sb * (t_air_k_i) ** 4
        - emiss_bare * c.sb * (lst) ** 4
    )

    if isinstance(rn_bare, xr.DataArray):
        rn_bare = rn_bare.transpose("time", "y", "x").chunk("auto")

    return rn_bare


def net_radiation_full(ra_hor_clear_i, emiss_atm_i, t_air_k_i, lst, r0_full=0.18):
    r"""
    Computes the net radiation at full canopy with zero evaporation.

    .. math ::

        Q_{full}^{*}=\left(1-\alpha_{0,full}\right) \cdot S_{d}+\varepsilon_{c} \cdot \varepsilon_{a} \cdot \sigma \cdot T_{a}^{4}-\varepsilon_{c} \cdot \sigma \cdot T_{s}^{4}

    Parameters
    ----------
    ra_hor_clear_i : float
        Total clear-sky irradiance on a horizontal surface, 
        :math:`ra_hor_clear_i`
        [W/m2]
    emiss_atm_i : float
        instantaneous atmospheric emissivity, 
        :math:`P`
        [-]
    t_air_k_i : float
        instantaneous air temperature, 
        :math:`T_{a}`
        [K]
    lst : float
        surface temperature, 
        :math:`T_{0}`
        [K]
    r0_full : float
        surface albedo full vegetation, 
        :math:`\alpha_{0, full}`
        [-]

    Returns
    -------
    rn_full : float
        net radiation full vegetation, 
        :math:`Q^*_{full}`
        [Wm-2]
    """
    emiss_full = 0.99
    rn_full = (
        (1 - r0_full) * ra_hor_clear_i
        + emiss_atm_i * emiss_full * c.sb * (t_air_k_i) ** 4
        - emiss_full * c.sb * (lst) ** 4
    )

    if isinstance(rn_full, xr.DataArray):
        rn_full = rn_full.transpose("time", "y", "x").chunk("auto")

    return rn_full


def sensible_heat_flux_bare(rn_bare, fraction_h_bare=0.65):
    r"""
    Computes the bare soil sensible heat flux.

    .. math ::

        H_{bare} = H_{f, bare} \cdot Q^*_{bare}

    Parameters
    ----------
    rn_bare : float
        net radiation bare soil, 
        :math:`Q^*_{bare}`
        [Wm-2]
    fraction_h_bare : float
        fraction of H of net radiation bare soil, 
        :math:`H_{f, bare}`
        [-]

    Returns
    -------
    h_bare : float
        sensible heat flux bare soil, 
        :math:`H_{bare}`
        [Wm-2]
     """
    return rn_bare * fraction_h_bare


def sensible_heat_flux_full(rn_full, fraction_h_full=0.95):
    r"""
    Computes the full canopy sensible heat flux.

    .. math ::

        H_{full} = H_{f, full} \cdot Q^*_{full}

    Parameters
    ----------
    rn_full : float
        net radiation full vegetation, 
        :math:`Q^*_{full}`
        [Wm-2]
    fraction_h_full : float
        fraction of H of net radiation full vegetation, 
        :math:`H_{f, full}`
        [-]

    Returns
    -------
    h_full : float
        sensible heat flux full vegetation, 
        :math:`H_{full}`
        [Wm-2]
     """
    return rn_full * fraction_h_full


def wind_speed_blending_height_bare(u_i, z0m_bare=0.001, z_obs=10, z_b=100):
    r"""
    Computes the wind speed at blending height :math:`u_{b}` [m/s] using the
    logarithmic wind profile.

    .. math ::

        u_{b}=\frac{u_{obs} \cdot \ln\left(\frac{z_{b}}{z_{0,m}}\right)}
        {\ln\left(\frac{z_{obs}}{z_{0,m}}\right)}

    Parameters
    ----------
    u_i : float
        instantaneous wind speed at observation height, 
        :math:`u_{obs}`
        [m/s]
    z_obs : float
        observation height of wind speed, 
        :math:`z_{obs}`
        [m]
    z_b : float
        blending height, 
        :math:`z_{b}`
        [m]
    z0m_bare : float
        surface roughness bare soil, 
        :math:`z_{0,m}`
        m

    Returns
    -------
    u_b_i_bare : float
        instantaneous wind speed at blending height for bare soil, 
        :math:`u_{b,i,bare}`
        [m/s]
    """
    ws = (c.k * u_i) / np.log(z_obs / z0m_bare) * np.log(z_b / z0m_bare) / c.k

    if isinstance(ws, xr.DataArray):
        ws = ws.clip(1, 150)
    else:
        ws = np.clip(ws, 1, 150)

    return ws


def wind_speed_blending_height_full_inst(u_i, z0m_full=0.1, z_obs=10, z_b=100):
    r"""
    Computes the wind speed at blending height :math:`u_{b}` [m/s] using the
    logarithmic wind profile.

    .. math ::

        u_{b}=\frac{u_{obs} \cdot \ln\left(\frac{z_{b}}{z_{0,m}}\right)}
        {\ln\left(\frac{z_{obs}}{z_{0,m}}\right)}

    Parameters
    ----------
    u_i : float
        instantaneous wind speed at observation height, 
        :math:`u_{obs}`
        [m/s]
    z_obs : float
        observation height of wind speed, 
        :math:`z_{obs}`
        [m]
    z_b : float
        blending height, 
        :math:`z_{b}`
        [m]
    z0m_full : float
        surface roughness vegetation, 
        :math:`z_{0,m}`
        [m]

    Returns
    -------
    u_b_i_full : float
        instantaneous wind speed at blending height for full vegetation, 
        :math:`u_{b,i,full}`
        [m s-1]
    """
    ws = (c.k * u_i) / np.log(z_obs / z0m_full) * np.log(z_b / z0m_full) / c.k

    if isinstance(ws, xr.DataArray):
        ws = ws.clip(1, 150)
    else:
        ws = np.clip(ws, 1, 150)

    return ws


def friction_velocity_full_inst(u_b_i_full, z0m_full=0.1, disp_full=0.667, z_b=100):
    r"""
    Like :func:`initial_friction_velocity_inst` but with full vegetation parameters.

    Parameters
    ----------
    u_b_i_full : float
        instantaneous wind speed blending height for full vegetation, 
        :math:`u_{b,d}`
        [m s-1]
    z0m_full : float
        surface roughness vegetation, 
        :math:`z_{0,m,b}`
        [m]
    disp_full : float
        displacement height vegetation, 
        :math:`d^{b}`
        [m]
    z_b : float
        blending height, 
        :math:`z_b`
        [m]

    Returns
    -------
    u_star_i_full : float
        instantaneous friction velocity vegetation, 
        :math:`u_{f}^{*}`
        [m s-1]

    """
    return initial_friction_velocity_inst(u_b_i_full, z0m_full, disp_full, z_b=100)


def friction_velocity_bare_inst(u_b_i_bare, z0m_bare=0.001, disp_bare=0.0, z_b=100):
    r"""
    Like :func:`initial_friction_velocity_inst` but with bare soil parameters.

    Parameters
    ----------
    u_b_i_bare : float
        instantaneous wind speed blending height bare soil, 
        :math:`u_{b,d}`
        [W m-2]
    z0m_bare : float
        surface roughness bare soil, 
        :math:`z_{0,m,b}`
        [m]
    disp_bare : float
        displacement height bare soil, 
        :math:`d^{b}`
        [m]
    z_b : float
        blending height, 
        :math:`z_b`
        [m]

    Returns
    -------
    u_star_i_bare : float
        instantaneous friction velocity bare soil, 
        :math:`u_{b}^{*}`
        [m s-1]

    """
    return initial_friction_velocity_inst(u_b_i_bare, z0m_bare, disp_bare, z_b=100)


def monin_obukhov_length_bare(h_bare, ad_i, u_star_i_bare, t_air_k_i):
    r"""
    Like :func:`unstable.monin_obukhov_length` but with bare soil parameters.

    Parameters
    ----------
    h_bare : float
        sensible heat flux for dry bare soil, 
        :math:`H_{b,d}`
        [W m-2]
    ad_i : float
        instantaneous air density, 
        :math:`\rho`
        [k g m-3]
    u_star_i_bare : float
        instantaneous friction velocity bare soil, 
        :math:`u^{*}_{b}`
        [m s-1]
    t_air_k_i : float
        instantaneous air temperature, 
        :math:`T_{a}`
        [K]

    Returns
    -------
    L_bare : float
        monin obukhov length dry vegetation, 
        :math:`L_{b,d}`
        [m]

    """
    return unstable.monin_obukhov_length(h_bare, ad_i, u_star_i_bare, t_air_k_i)


def monin_obukhov_length_full(h_full, ad_i, u_star_i_full, t_air_k_i):
    r"""
    Like :func:`unstable.monin_obukhov_length` but with full canopy parameters.

    Parameters
    ----------
    h_full : float
        sensible heat flux for dry full vegetation, 
        :math:`H_{f,d}`
        [W m-2]
    ad_i : float
        instantaneous air density, 
        :math:`\rho`
        [k g m-3]
    u_star_i_full : float
        instantaneous friction velocity vegetation, 
        :math:`u^{*}_{b}`
        [m s-1]
    t_air_k_i : float
        instantaneous air temperature, 
        :math:`T_{a}`
        [K]

    Returns
    -------
    L_full : float
        monin obukhov length dry vegetation, 
        :math:`L_{f,d}`
        [m]

    """
    return unstable.monin_obukhov_length(h_full, ad_i, u_star_i_full, t_air_k_i)


def aerodynamical_resistance_forced_convection_full(u_i, L_full, z0m_full=0.1, disp_full=0.667, z_obs=10):
    r"""
    Computes the aerodynamical resistance for a full canopy, Eq A1 from Sanchez.

    .. math ::

        z_{1} &= \frac{z_{obs}-d}{z_{0,m}} \\
        z_{2} &= \frac{z_{obs}-d}{L} \\
        z_{3} &= \frac{z_{0,m}}{L} \\
        z_{4} &= \frac{z_{obs}-d}{\frac{z_{0,m}}{7}} \\
        z_{5} &= \frac{\frac{z_{0,m}}{7}}{L} \\
        r_{a,c} &= \frac{(\ln(z_{1})-\phi_{m}(-z_{2})+\phi_{m}(-z_{3})) \cdot (\ln(z_{4})-\phi_{h}(-z_{2})+\phi_{h}(-z_{5}))}{k^{2} \cdot u}

    Parameters
    ----------
    u_i : float
        instantaneous wind speed at observation height, 
        :math:`u_{obs}`
        [m/s]
    z_obs : float
        observation height of wind speed, 
        :math:`z_{obs}`
        [m]
    disp_full : float
        displacement height, 
        :math:`d`
        [m]
    z0m_full : float
        surface roughness, 
        :math:`z_{0,m}`
        [m]
    L_full : float
        monin obukhov length, 
        :math:`L`
        [m]

    Returns
    -------
    rac : float
        aerodynamical resistance canopy, 
        :math:`r_{a,c}`
        [sm-1]

    """
    z1 = (z_obs - disp_full) / z0m_full
    z2 = (z_obs - disp_full) / L_full
    z3 = z0m_full / L_full
    z4 = (z_obs - disp_full) / (z0m_full / 7)
    z5 = (z0m_full / 7) / L_full
    res = (
        (np.log(z1) - psi_m(-z2) + psi_m(-z3)) * (np.log(z4) - psi_h(-z2) + psi_h(-z5))
    ) / (c.k ** 2 * u_i)

    return res


def aerodynamical_resistance_forced_convection_bare(u_i, L_bare, z0m_bare=0.001, disp_bare=0.0, z_obs=10):
    r"""
    Computes the aerodynamical resistance for a dry bare soil.

    .. math ::

        z_{1} &= \frac{z_{obs}-d}{z_{0,b,m}} \\
        z_{2} &= \frac{z_{obs}-d}{L_{b}} \\
        r_{a,a} &= \frac{(\ln(z_{1})-\phi_{m}(-z_{2})) \cdot (\ln(z_{1})-\phi_{h}(-z_{2}))}{k^{2} \cdot u} \\

    Parameters
    ----------
    u_i : float
        instantaneous wind speed at observation height, 
        :math:`u_{obs}`
        [m/s]
    z_obs : float
        observation height of wind speed, 
        :math:`z_{obs}`
        [m]
    disp_bare : float
        displacement height, 
        :math:`d`
        [m]
    z0m_bare : float
        surface roughness, 
        :math:`z_{0,b,m}`
        [m]
    L_bare : float
        monin obukhov length, 
        :math:`L_{b}`
        [m]

    Returns
    -------
    raa : float
        aerodynamical resistance dry surface, 
        :math:`r_{a,a}`
        [sm-1]

    """

    z1 = (z_obs - disp_bare) / z0m_bare
    z2 = (z_obs - disp_bare) / L_bare
    res = ((np.log(z1) - psi_m(-z2)) * (np.log(z1) - psi_h(-z2))) / (c.k ** 2 * u_i)

    return res


def wind_speed_soil_inst(u_i, L_bare, z_obs=10):
    r"""
    Computes the instantaneous wind speed at soil surface.

    .. math ::

        u_{i,s}=u_{obs} \cdot \frac{\ln\left(\frac{z_{obs}}{z_{0}}\right)}
              {\ln\left(\frac{z_{obs}}{z_{0,s}}\right)-\psi_{m}\left(\frac{-z_{0}}{L}\right)}

    Parameters
    ----------
    u_i : float
        wind speed at observation height, 
        :math:`u_{obs}`
        [m/s]
    z_obs : float
        observation height of wind speed, 
        :math:`z_{obs}`
        [m]
    L_bare : float
        monin obukhov length, 
        :math:`L`
        [m]

    Returns
    -------
    u_i_soil : float
        instantaneous wind speed just above soil surface, 
        :math:`u_{i,s}`
        [ms-1]

    """
    z0_soil = 0.01
    z0_free = 0.1
    return u_i * (
        (np.log(z0_free / z0_soil))
        / (np.log(z_obs / z0_soil) - psi_m(-z0_free / L_bare))
    )


def maximum_temperature_full(
    ra_hor_clear_i, emiss_atm_i, t_air_k_i, ad_i, rac, r0_full=0.18
):
    r"""
    Computes the maximum temperature under fully vegetated conditions.

    .. math ::

        T_{c,max}=\frac{\left(1-\alpha_{c}\right) S_{d}+\varepsilon_{c}\varepsilon_{a}\sigma
                  T_{a}^{4}-\varepsilon_{c}\sigma T_{a}^{4}}{4\varepsilon_{s}\sigma
                  T_{a}^{3}+\rho C_{p}/r_{a,c}}+T_{a}

    Parameters
    ----------
    ra_hor_clear_i : float
        Total clear-sky irradiance on a horizontal surface, 
        :math:`ra_hor_clear_i`
        [W/m2]
    emiss_atm_i : float
        instantaneous atmospheric emissivity, 
        :math:`P`
        [-]
    t_air_k_i : float
        instantaneous air temperature, 
        :math:`T_{a}`
        [K]
    rac : float
        aerodynamic resistance canopy, 
        :math:`r_{a,c}`
        [sm-1]
    ad_i : float
        instantaneous air density, 
        :math:`\rho`
        [kg m-3]
    r0_full : float
        surface albedo full vegetation cover, 
        :math:`\alpha_{0, full}`
        [-]

    Returns
    -------
    t_max_full : float
        maximum temperature at full vegetation cover, 
        :math:`T_{c,max}`
        [K]

    """
    emiss_full = 0.99

    tc_max_num = (
        (1 - r0_full) * ra_hor_clear_i
        + emiss_full * emiss_atm_i * c.sb * (t_air_k_i) ** 4
        - emiss_full * c.sb * (t_air_k_i) ** 4
    )
    tc_max_denom = 4 * emiss_full * c.sb * (t_air_k_i) ** 3 + (ad_i * c.sh) / rac
    tc_max = tc_max_num / tc_max_denom + t_air_k_i

    if isinstance(tc_max, xr.DataArray):
        tc_max = tc_max.transpose("time", "y", "x").chunk("auto")

    return tc_max


def maximum_temperature_bare(
    ra_hor_clear_i, emiss_atm_i, t_air_k_i, ad_i, raa, ras, r0_bare=0.38
):
    r"""
    Computes the maximum temperature under dry bare soil conditions.

    .. math ::

        T_{s,max}=\frac{\left(1-\alpha_{s}\right)S_{d}+\varepsilon_{s}\varepsilon_{a}\sigma
        T_{a}^{4}-\varepsilon_{s}\sigma T_{a}^{4}}{4\varepsilon_{s}\sigma T_{a}^{3}+
        \rho C_{p}/\left[\left(r_{a,a}+r_{a,s}\right)\left(1-G/R_{n,s}\right)\right]}+T_{a}

    Parameters
    ----------
    ra_hor_clear_i : float
        Total clear-sky irradiance on a horizontal surface, 
        :math:`ra_hor_clear_i`
        [W/m2]
    emiss_atm_i : float
        instantaneous atmospheric emissivity, 
        :math:`P`
        [-]
    t_air_k_i : float
        instantaneous air temperature, 
        :math:`T_{a}`
        [K]
    ad_i : float
        instantaneous air density, 
        :math:`\rho`
        [kg m-3]
    raa : float
        aerodynamical resistance, 
        :math:`r_{a,a}`
        [sm-1]
    ras : float
        aerodynamical resistance, 
        :math:`r_{a,s}`
        [sm-1]
    r0_bare : float
        dry bare soil surface albedo, 
        :math:`\alpha_{0, bare}`
        [-]

    Returns
    -------
    t_max_bare : float
        maximum temperature at bare soil, 
        :math:`T_{c,max}`
        [K]

    """
    emiss_bare = 0.95
    ts_max_num = (
        (1 - r0_bare) * ra_hor_clear_i
        + emiss_bare * emiss_atm_i * c.sb * (t_air_k_i) ** 4
        - emiss_bare * c.sb * (t_air_k_i) ** 4
    )
    ts_max_denom = 4 * emiss_bare * c.sb * (t_air_k_i) ** 3 + (ad_i * c.sh) / (
        (raa + ras) * (1 - 0.35)
    )

    out = ts_max_num / ts_max_denom + t_air_k_i

    if isinstance(out, xr.DataArray):
        out = out.transpose("time", "y", "x").chunk("auto")

    return out


def maximum_temperature(t_max_bare, t_max_full, vc):
    r"""
    Computes the maximum temperature at dry conditions.

    .. math ::

        T_{0,max} = c_{veg} \cdot (T_{c,max}-T_{s,max})+T_{s,max}


    Parameters
    ----------
    t_max_bare : float
        maximum temperature at bare soil, 
        :math:`T_{s,max}`
        [K]
    t_max_full : float
        maximum temperature at full dry vegetation, 
        :math:`T_{c,max}`
        [K]
    vc : float
        vegetation cover, 
        :math:`c_{veg}`
        [-]


    Returns
    -------
    lst_max : float
        maximum temperature at dry conditions, 
        :math:`T_{0,max}`
        [K]

    """
    out = vc * (t_max_full - t_max_bare) + t_max_bare

    if isinstance(out, xr.DataArray):
        out = out.transpose("time", "y", "x").chunk("auto")

    return out


def minimum_temperature(t_wet_k_i, t_air_k_i, vc):
    r"""
    Computes the maximum temperature at dry conditions.

    .. math ::

        T_{0,min} = c_{veg} \cdot (T_{a,i}-T_{w})+T_{w}


    Parameters
    ----------
    t_wet_k_i : float
        minimum temperature at bare soil, 
        :math:`T_{s,max}`
        [K]
    t_air_k_i : float
        minimum temperature at full vegetation, 
        :math:`T_{c,max}`
        [K]
    vc : float
        vegetation cover, 
        :math:`c_{veg}`
        [-]


    Returns
    -------
    lst_min : float
        minimum temperature at wet conditions, 
        :math:`T_{0,min}`
        [K]

    """
    out = vc * (t_air_k_i - t_wet_k_i) + t_wet_k_i

    if isinstance(out, xr.DataArray):
        out = out.transpose("time", "y", "x").chunk("auto")

    return out

def soil_moisture_from_maximum_temperature(lst_max, lst, lst_min):
    r"""
    Computes the relative root zone soil moisture based on estimates of
    maximum temperature and wet bulb temperature and measured land
    surface temperature.

    .. math ::

        \Theta = \frac{T_{0}-T_{0,min}}{T_{0,max}-T_{0,min}}

    Parameters
    ----------
    lst : float
        land surface temperature, 
        :math:`T_{0}`
        [K]
    lst_max : float
        maximum temperature at dry conditions, 
        :math:`T_{0,max}`
        [K]
    lst_min : float
        minimum temperature at wet conditions, 
        :math:`T_{0, min}`
        [K]


    Returns
    -------
    se_root : float
        relative root zone soil moisture (based on LST), 
        :math:`\Theta`
        [%]

    """
    ratio = (lst - lst_min) / (lst_max - lst_min)

    if isinstance(ratio, xr.DataArray):
        ratio = ratio.clip(0, 1)
    else:
        ratio = np.clip(ratio, 0, 1)

    return 1 - ratio

def aerodynamical_resistance_forced_convection_soil(u_i_soil):
    r"""
    Computes the aerodynamical resistance of the soil

    Eq A7 from Sanchez.

    .. math ::
        r_{a,s}=\frac{1}{\left(0.0025T_{dif}^{\frac{1}{3}}+0.012u_{i,s}\right)}

    Parameters
    ----------
    u_i_soil : float
        instantaneous wind speed just above soil surface
        :math:`u_{i,s}`
        [m s-1]

    Returns
    -------
    ras_forced : float
        forced aerodynamical resistance
        :math:`r_{a,s}`
        [sm-1]

    """
    Tdif = 10.0
    return 1. / (0.0025 * (Tdif) ** (1. / 3.) + 0.012 * u_i_soil)


def aerodynamical_resistance_free_convection_bare(h_bare, t_air_k_i, ad_i, z0m_bare=0.001):
    r"""
    Calculates the aerodynamic resistance in the case of free convection for bare soil.

    Based on equation 15 from:
    https://journals.ametsoc.org/view/journals/apme/31/9/1520-0450_1992_031_1096_emlst_2_0_co_2.xml

    Parameters
    ----------
    h_bare : float
        sensible heat flux bare soil
        :math:`H_{bare}`
        [Wm-2]
    t_air_k_i : float
        instantaneous air temperature
        :math:`T_{a}`
        [K]
    ad_i : float
        instantaneous air density
        :math:`\rho`
        [k g m-3]
    z0m_bare : float
        surface roughness bare soil
        :math:`z_{0,m}`
        [m]

    Returns
    -------
    rah_bare_free : float
        aerodynamic resistance soil
        :math:`r_{a,a}`
        [sm-1]
    """

    # approximation: tpot = tact
    θ = t_air_k_i
    z0h_bare = z0m_bare / 7.

    rah_bare_free = 1.00434 / (np.cbrt(z0h_bare/θ) * np.cbrt(h_bare/(ad_i*c.sh)))

    return rah_bare_free


def aerodynamical_resistance_free_convection_full(h_full, t_air_k_i, ad_i, z0m_full=0.1):
    r"""
    Calculates the aerodynamic resistance in the case of free convection for the canopy.

    Based on equation 15 from:
    https://journals.ametsoc.org/view/journals/apme/31/9/1520-0450_1992_031_1096_emlst_2_0_co_2.xml

    Parameters
    ----------
    h_full : float
        sensible heat flux full canopy
        :math:`H_{full}`
        [Wm-2]
    t_air_k_i : float
        instantaneous air temperature
        :math:`T_{a}`
        [K]
    ad_i : float
        instantaneous air density
        :math:`\rho`
        [k g m-3]
    z0m_full : float
        surface roughness full canopy
        :math:`z_{0,m}`
        [m]

    Returns
    -------
    rah_full_free : float
        aerodynamic resistance canopy
        :math:`r_{a,c}`
        [sm-1]
    """

    # approximation: tpot = tact
    θ = t_air_k_i
    z0h_full = z0m_full / 7.

    rah_full_free = 1.00434 / (np.cbrt(z0h_full/θ) * np.cbrt(h_full/(ad_i*c.sh)))

    return rah_full_free


def aerodynamical_resistance_bare(raa_forced, ras_forced, rah_bare_free=np.inf):
    r"""
    Computes the aerodynamical resistance for a dry bare soil.

    This function takes the minimum of forced and free convection

    Parameters
    ----------
    raa_forced : float
        forced aerodynamical resistance dry surface
        :math:`r_{a,a}`
        [sm-1]
    ras_forced : float
        forced aerodynamical resistance
        :math:`r_{a,s}`
        [sm-1]
    rah_bare_free : float
        aerodynamic resistance soil
        :math:`r_{a,a}`
        [sm-1]

    Returns
    -------
    rah_bare : float
        aerodynamical resistance bare soil
        :math:`r_{a,h} bare`
        [sm-1]
    """

    # soil forced resistance in series
    rah_bare = np.minimum(raa_forced + ras_forced, rah_bare_free)

    return rah_bare


def aerodynamical_resistance_full(rac_forced, rah_full_free=np.inf):
    r"""
    Computes the aerodynamical resistance for a full canopy.

    This function takes the minimum of forced and free convection

    Parameters
    ----------
    rac_forced : float
        forced aerodynamical resistance canopy
        :math:`r_{a,c}`
        [sm-1]
    rah_full_free : float
        aerodynamic resistance soil
        :math:`r_{a,a}`
        [sm-1]

    Returns
    -------
    rah_full : float
        aerodynamical resistance canopy
        :math:`r_{a,h} canopy`
        [sm-1]

    """

    rah_full = np.minimum(rac_forced, rah_full_free)

    return rah_full


import os
from pywapor.collect.protocol import cog
from pywapor.general.processing_functions import open_ds, remove_ds, save_ds
import xarray as xr
import numpy as np
import copy

def default_vars(product_name, req_vars):
    """Given a `product_name` and a list of requested variables, returns a dictionary
    with metadata on which exact layers need to be requested from the server, how they should
    be renamed, and how their dimensions are defined.

    Parameters
    ----------
    product_name : str
        Name of the product.
    req_vars : list
        List of variables to be collected.

    Returns
    -------
    dict
        Metadata on which exact layers need to be requested from the server.
    """
    variables = {
        'WaPOR2': {
                "Band1": [("lat", "lon"), "z_oro"],
                "Band2": [("lat", "lon"), "vpd_slope"],
                "Band3": [("lat", "lon"), "t_amp"],
                "Band4": [("lat", "lon"), "rn_offset"],
                "Band5": [("lat", "lon"), "r0_full"],
                "Band6": [("lat", "lon"), "r0_bare"],
                "Band7": [("lat", "lon"), "rs_min"],
                "Band8": [("lat", "lon"), "rn_slope"],
                "Band9": [("lat", "lon"), "t_opt"],
                "Band10": [("lat", "lon"), "lw_slope"],
                "Band11": [("lat", "lon"), "land_mask"],
                "Band12": [("lat", "lon"), "lw_offset"],
                "Band13": [("lat", "lon"), "z_obst_max"],
                "crs": [(), "spatial_ref"],
                    },
        # 'WaPOR3': {
        #         "Band2": [("lat", "lon"), "lw_offset"],
        #         "Band1": [("lat", "lon"), "lw_slope"],
        #         "crs": [(), "spatial_ref"],
        # },
        'WaPOR3': {
            "Band1": [("lat", "lon"), "lw_offset"],
            "Band2": [("lat", "lon"), "lw_slope"],
            "Band3": [("lat", "lon"), "rn_offset"],
            "Band4": [("lat", "lon"), "rn_slope"],
            "Band5": [("lat", "lon"), "t_amp"],
            "Band6": [("lat", "lon"), "t_opt"],
            "Band7": [("lat", "lon"), "vpd_slope"],
            "crs": [(), "spatial_ref"],
        }
    }

    req_dl_vars = {
        "WaPOR2": {
            "z_oro": ["Band1", "crs"],
            "vpd_slope": ["Band2", "crs"],
            "t_amp": ["Band3", "crs"],
            "rn_offset": ["Band4", "crs"],
            "r0_full": ["Band5", "crs"],
            "r0_bare": ["Band6", "crs"],
            "rs_min": ["Band7", "crs"],
            "rn_slope": ["Band8", "crs"],
            "t_opt": ["Band9", "crs"],
            "lw_slope": ["Band10", "crs"],
            "land_mask": ["Band11", "crs"],
            "lw_offset": ["Band12", "crs"],
            "z_obst_max": ["Band13", "crs"],
        },
        # "WaPOR3": {
        #     "lw_offset": ["Band2", "crs"],
        #     "lw_slope": ["Band1", "crs"],
        # },
        "WaPOR3": {
            "lw_offset": ["Band1", "crs"], 
            "lw_slope": ["Band2", "crs"], 
            "rn_offset": ["Band3", "crs"], 
            "rn_slope": ["Band4", "crs"], 
            "t_amp": ["Band5", "crs"], 
            "t_opt": ["Band6", "crs"], 
            "vpd_slope": ["Band7", "crs"], 
        }
    }

    out = {val:variables[product_name][val] for sublist in map(req_dl_vars[product_name].get, req_vars) for val in sublist}
    
    return out

def scale_factor(ds, var, scale = 0.01):
    ds[var] = ds[var] * scale
    return ds

def default_post_processors(product_name, req_vars):
    """Given a `product_name` and a list of requested variables, returns a dictionary with a 
    list of functions per variable that should be applied after having collected the data
    from a server.

    Parameters
    ----------
    product_name : str
        Name of the product.
    req_vars : list
        List of variables to be collected.

    Returns
    -------
    dict
        Functions per variable that should be applied to the variable.
    """
    post_processors = {
        'WaPOR2': {
                    "land_mask": [],
                    "lw_offset": [],
                    "lw_slope": [],
                    "r0_bare": [],
                    "r0_full": [],
                    "rn_offset": [],
                    "rn_slope": [],
                    "rs_min": [],
                    "t_amp": [],
                    "t_opt": [],
                    "vpd_slope": [],
                    "z_obst_max": [],
                    "z_oro": [],
                    },
        # 'WaPOR3': {
        #             "lw_offset": [],
        #             "lw_slope": [],
        #             },
        "WaPOR3": {
            "lw_offset": [], 
            "lw_slope": [], 
            "rn_offset": [], 
            "rn_slope": [], 
            "t_amp": [], 
            "t_opt": [], 
            "vpd_slope": [], 
        }
    }

    out = {k:v for k,v in post_processors[product_name].items() if k in req_vars}

    return out

def url_func(product_name):
    """Returns a url at which to collect MERRA2 data.

    Parameters
    ----------
    product_name : None
        Not used.

    Returns
    -------
    str
        The url.
    """
    url = f"https://storage.googleapis.com/fao-cog-data/{product_name}.tif"
    return url

def download(folder, latlim, lonlim, product_name, req_vars,
                variables = None, post_processors = None, **kwargs):
    """Download MODIS data and store it in a single netCDF file.

    Parameters
    ----------
    folder : str
        Path to folder in which to store results.
    latlim : list
        Latitude limits of area of interest.
    lonlim : list
        Longitude limits of area of interest.
    timelim : list
        Period for which to prepare data.
    product_name : str
        Name of the product to download.
    req_vars : list
        Which variables to download for the selected product.
    variables : dict, optional
        Metadata on which exact layers need to be requested from the server, by default None.
    post_processors : dict, optional
        Functions per variable that should be applied to the variable, by default None.

    Returns
    -------
    xr.Dataset
        Downloaded data.
    """
    folder = os.path.join(folder, "STATICS")
    
    fn = os.path.join(folder, f"{product_name}.nc")
    req_vars_orig = copy.deepcopy(req_vars)
    if os.path.isfile(fn):
        existing_ds = open_ds(fn)
        req_vars_new = list(set(req_vars).difference(set(existing_ds.data_vars)))
        if len(req_vars_new) > 0:
            req_vars = req_vars_new
            existing_ds = existing_ds.close()
        else:
            return existing_ds[req_vars_orig]

    spatial_buffer = True
    if spatial_buffer:
        dx = dy = 0.2
        latlim = [latlim[0] - dy, latlim[1] + dy]
        lonlim = [lonlim[0] - dx, lonlim[1] + dx]

    coords = {"x": ("lon", lonlim), "y": ("lat", latlim)}

    if isinstance(variables, type(None)):
        variables = default_vars(product_name, req_vars)

    if isinstance(post_processors, type(None)):
        post_processors = default_post_processors(product_name, req_vars)
    else:
        default_processors = default_post_processors(product_name, req_vars)
        post_processors = {k: {True: default_processors[k], False: v}[v == "default"] for k,v in post_processors.items() if k in req_vars}

    ds = cog.download(fn, product_name, coords, variables, 
                        post_processors, url_func)

    return ds[req_vars_orig]

if __name__ == "__main__":

    req_vars = [
                # 'land_mask',
                'lw_offset',
                'lw_slope',
                # 'r0_bare',
                # 'r0_full',
                'rn_offset',
                'rn_slope',
                # 'rs_min',
                # 't_amp',
                't_amp',
                't_opt',
                'vpd_slope',
                # 'z_obst_max',
                # 'z_oro'
                ]

    product_name = "WaPOR3"

    folder = r"/Users/hmcoerver/Local/new_statics"
    # latlim = [28.9, 29.7]
    # lonlim = [30.2, 31.2]
    latlim = [21.9692194682626933, 21.9939120838340507]
    lonlim = [91.9371349243682801, 91.9657566608824339]

    variables = None
    post_processors = None

    ds = download(folder, latlim, lonlim, product_name, req_vars,
                variables = None, post_processors = None)



"""
===========================================
 pywapor.et_look_v2 functions (:mod:` pywapor.et_look_v2`)
===========================================

.. currentmodule::  pywapor.et_look_v2

Within the :mod:` pywapor.et_look_v2` module all physical and empirical
functions related to the calculation of the soil moisture, interceptin,
evaporation and transpiration are provided. These functions listed here
can be used to build function chains.

Instantaneous Radiation (:mod:` pywapor.et_look_v2.clear_sky_radiation`)
==============================================================
.. automodule::  pywapor.et_look_v2.clear_sky_radiation

.. autosummary::
   :toctree: generated/

   beam_irradiance_horizontal_clear
   beam_irradiance_normal_clear
   day_angle
   declination
   diffuse_irradiance_horizontal_clear
   extraterrestrial_irradiance_normal
   hour_angle
   inverse_earth_sun_distance
   linke_turbidity
   ra_clear_horizontal
   rayleigh_optical_thickness
   relative_optical_airmass
   solar_constant
   solar_elevation_angle
   solar_elevation_angle_refracted

Evapotranspiration (:mod:` pywapor.et_look_v2.evapotranspiration`)
===========================================================================

.. automodule::  pywapor.et_look_v2.evapotranspiration

.. autosummary::
    :toctree: generated/

    et_actual_mm
    et_reference
    et_reference_mm
    interception_mm

Vegetation Cover (:mod:` pywapor.et_look_v2.leaf`)
===========================================================

.. automodule::  pywapor.et_look_v2.leaf

.. autosummary::
   :toctree: generated/

   vegetation_cover
   leaf_area_index
   effective_leaf_area_index

Meteorology (:mod:` pywapor.et_look_v2.meteo`)
==========================================================

.. automodule::  pywapor.et_look_v2.meteo

.. autosummary::
   :toctree: generated/

   air_density
   air_density_daily
   air_density_inst
   air_pressure
   air_pressure_daily
   air_temperature_kelvin
   air_temperature_kelvin_daily
   air_temperature_kelvin_inst
   disaggregate_air_temperature
   disaggregate_air_temperature_daily
   disaggregate_air_temperature_inst
   disaggregate_dew_point_temperature_inst
   dry_air_density
   dry_air_density_daily
   dry_air_density_inst
   latent_heat
   latent_heat_daily
   moist_air_density
   moist_air_density_daily
   moist_air_density_inst
   psychrometric_constant
   psychrometric_constant_daily
   saturated_vapour_pressure
   saturated_vapour_pressure_minimum
   saturated_vapour_pressure_maximum
   saturated_vapour_pressure_average
   saturated_vapour_pressure_daily
   slope_saturated_vapour_pressure
   slope_saturated_vapour_pressure_daily
   vapour_pressure_deficit
   vapour_pressure_deficit_daily
   vapour_pressure_from_specific_humidity
   vapour_pressure_from_specific_humidity_daily
   vapour_pressure_from_specific_humidity_inst
   wet_bulb_temperature_kelvin_inst
   wind_speed_blending_height
   wind_speed_blending_height_daily



Net Available Energy (:mod:` pywapor.et_look_v2.radiation`)
=================================================================

.. automodule::  pywapor.et_look_v2.radiation

.. autosummary::
    :toctree: generated/

    bare_soil_heat_flux
    damping_depth
    interception_wm2
    longwave_radiation_fao
    longwave_radiation_fao_etref
    net_radiation
    net_radiation_canopy
    net_radiation_grass
    net_radiation_soil
    soil_fraction
    soil_heat_flux
    soil_thermal_conductivity
    volumetric_heat_capacity

Roughness (:mod:` pywapor.et_look_v2.roughness`)
=============================================

.. automodule::  pywapor.et_look_v2.roughness

.. autosummary::
   :toctree: generated/

   roughness_length
   obstacle_height
   displacement_height

Solar radiation (:mod:` pywapor.et_look_v2.solar_radiation`)
=================================================================

.. automodule::  pywapor.et_look_v2.solar_radiation

.. autosummary::
    :toctree: generated/

    aspect_rad
    cosine_solar_zenith_angle
    daily_solar_radiation_flat
    daily_solar_radiation_toa
    daily_solar_radiation_toa_flat
    daily_total_solar_radiation
    declination
    diffusion_index
    hour_angle
    inst_solar_radiation_toa
    inverse_earth_sun_distance
    latitude_rad
    seasonal_correction
    slope_rad
    sunset_hour_angle

Plant stress (:mod:` pywapor.et_look_v2.stress`)
===================================================

.. automodule::  pywapor.et_look_v2.stress

.. autosummary::
    :toctree: generated/

    stress_moisture
    stress_radiation
    stress_temperature
    stress_vpd

Canopy and Soil Resistance (:mod:` pywapor.et_look_v2.resistance`)
===========================================================

.. automodule::  pywapor.et_look_v2.resistance

.. autosummary::
    :toctree: generated/

    atmospheric_canopy_resistance
    canopy_resistance
    soil_resistance

Neutral Atmosphere (:mod:` pywapor.et_look_v2.neutral`)
=====================================================

.. automodule::  pywapor.et_look_v2.neutral

.. autosummary::
    :toctree: generated/

    initial_canopy_aerodynamic_resistance
    initial_daily_evaporation
    initial_daily_evaporation_mm
    initial_daily_transpiration
    initial_daily_transpiration_mm
    initial_soil_aerodynamic_resistance

Unstable Atmosphere (:mod:` pywapor.et_look_v2.unstable`)
=======================================================

.. automodule::  pywapor.et_look_v2.unstable

.. autosummary::
    :toctree: generated/

    evaporation
    evaporation_mm
    friction_velocity
    initial_friction_velocity_daily
    initial_sensible_heat_flux_canopy_daily
    initial_sensible_heat_flux_soil_daily
    monin_obukhov_length
    ra_canopy
    ra_soil
    stability_correction_heat_obs
    stability_factor
    stability_parameter
    stability_parameter_obs
    transpiration
    transpiration_mm


Soil Moisture (:mod:` pywapor.et_look_v2.soil_moisture`)
====================================================================

.. automodule::  pywapor.et_look_v2.soil_moisture

.. autosummary::
    :toctree: generated/

    psi_m
    psi_h
    aerodynamical_resistance_bare
    aerodynamical_resistance_full
    aerodynamical_resistance_soil
    atmospheric_emissivity_inst
    friction_velocity_bare_inst
    friction_velocity_full_inst
    initial_friction_velocity_inst
    minimum_temperature
    maximum_temperature
    maximum_temperature_bare
    maximum_temperature_full
    monin_obukhov_length_bare
    monin_obukhov_length_full
    net_radiation_bare
    net_radiation_full
    sensible_heat_flux_bare
    sensible_heat_flux_full
    soil_moisture_from_maximum_temperature
    wind_speed_blending_height_bare
    wind_speed_blending_height_full_inst
    wind_speed_soil_inst


"""
from pywapor.et_look_v2_v3 import solar_radiation, clear_sky_radiation, meteo, radiation, evapotranspiration, soil_moisture, leaf, stress, resistance, roughness, neutral, unstable, constants, biomass

__all__ = ['solar_radiation', 'clear_sky_radiation', 'meteo', 'radiation', 'evapotranspiration', 'soil_moisture', 'leaf', 'stress', 'resistance', 'roughness', 'neutral', 'unstable', 'constants', 'biomass']

__version__ = '0.1'



import numpy as np
import os
from osgeo import gdal
from collections import OrderedDict
from pywapor.general.logger import log

def curvi_to_recto(lats, lons, data, out_fn, warp_kwargs = {}):

    if isinstance(lats, str):
        lats = gdal.Open(lats)
    if isinstance(lons, str):
        lons = gdal.Open(lons)

    for name, x in data.items():
        if isinstance(x, str):
            data[name] = gdal.Open(x)

    if lats.RasterCount != 1:
        raise ValueError
    if lons.RasterCount != 1:
        raise ValueError
    
    xsize = lons.RasterXSize
    ysize = lons.RasterYSize
    if xsize != lats.RasterXSize:
        raise ValueError
    if ysize != lons.RasterYSize:
        raise ValueError

    data_filtered = OrderedDict()
    for name, x in data.items():
        if x.RasterCount > 0:
            data_filtered[name] = x
        else:
            log.warning(f"No Raster found in {name} ({x.GetFileList()})")

    #####
    # Create VRT file linking all the required data.
    #####
    vrt_options = gdal.BuildVRTOptions(separate = True)
    data_bands = gdal.BuildVRT("/vsimem/temp.vrt", list(data_filtered.values()), options = vrt_options) 

    items = {"LINE_OFFSET": "0", 
            "LINE_STEP": "1", 
            "PIXEL_OFFSET": "0", 
            "PIXEL_STEP": "1", 
            "X_BAND": "1",
            "X_DATASET": lons.GetDescription(),
            "Y_BAND": "1",
            "Y_DATASET": lats.GetDescription()}

    _ = data_bands.SetMetadata(items, "GEOLOCATION")
    data_bands.FlushCache()

    #####
    # Warp the data.
    #####
    coptions = [
        "COMPRESS=DEFLATE"
        ]

    options = gdal.WarpOptions(
        srcSRS = "epsg:4326",
        dstSRS = "epsg:4326",
        dstNodata = -9999,
        geoloc = True,
        creationOptions = coptions,
        **warp_kwargs,
    )

    try:
        # NOTE turn of logging because of irrelevant ERROR message. "NUL" for windows, other for mac/linux.
        # also see https://gis.stackexchange.com/questions/358304/suppress-warnings-in-python-ogr
        gdal.SetConfigOption("CPL_LOG", {"nt": "NUL"}.get(os.name, "/dev/null"))
        out = gdal.Warp(out_fn, data_bands, options = options)
    except Exception as e:
        raise e
    finally:
        gdal.SetConfigOption("CPL_LOG", None)

    _ = [out.GetRasterBand(i + 1).SetMetadata({"variable": list(data_filtered.keys())[i]}) for i in range(out.RasterCount)]
    out.FlushCache()
    out = None

    for x, ds in data.items():
        ds.FlushCache()
        data[x] = None

    lats.FlushCache()
    lats = None
    lons.FlushCache()
    lons = None

    return out_fn

def create_grid(latlim, lonlim, dx_dy = (0.0033, 0.0033)):
    dx, dy = dx_dy
    nx = int(np.ceil((lonlim[1] - lonlim[0]) / dx))
    ny = int(np.ceil((latlim[1] - latlim[0]) / dy))
    bb = [lonlim[0], latlim[0], lonlim[0] + nx * dx, latlim[0] + ny * dy]
    return bb, nx, ny
